// Copyright (c) 2018 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.io.*;

public class DebugMemory implements Memory {
	RandomAccessFile core;
	byte[] mem;
	Z80Disassembler dasm = null;
	int org;
	long len;
	long end;
	long cur = -1;

	public DebugMemory(File img, int org) {
		this.org = org;
		loadCore(img);
	}

	public DebugMemory(File img, int org, boolean zilog) {
		this.org = org;
		loadCore(img);
		if (zilog) {
			dasm = new Z80DisassemblerZilog(this);
		} else {
			dasm = new Z80DisassemblerMAC80(this);
		}
	}

	private void loadCore(File img) {
		RandomAccessFile fi = null;
		try {
			fi = new RandomAccessFile(img, "r");
			mem = new byte[1024];	// a "reasonable" buffer?
			len = fi.length();
			end = len + org;
		} catch (Exception ee) {
			ee.printStackTrace();
			System.exit(1);
		}
		core = fi;
	}

	public int read(boolean rom, int bank, int address) {
		return read((bank << 16) | address);
	}
	// address includes 'org'
	public int read(int address) {
		long adr = address - org;
		long off = (adr & ~0x3ff);
		if (cur != off) {
			cur = off;
			try {
				core.seek(off);
				core.read(mem);
			} catch (Exception ee) {
				// TODO: passively ignore? or fatal error?
				Arrays.fill(mem, (byte)0);
				//ee.printStackTrace();
				//System.exit(1);
			}
		}
		return mem[(int)(adr - cur)] & 0xff;
	}
	public void write(int address, int value) {}
	public void reset() {}
	public void dumpCore(String file) {}

	public int size() { return (int)len; }

	// Address is physical address, convert to bank+virt to avoid
	// truncation in Z80Disassembler, then convert back in read().
	// Address includes 'org'
	public int disas(int adr) {
		if (dasm == null) {
			return 16;	// finish sooner...
		}
		String i = dasm.disas(false, (adr >> 16), adr);
		int n = dasm.instrLen();
		String s = String.format("%05x:", adr);
		int x;
		for (x = 0; x < n; ++x) {
			s += String.format(" %02x", read(adr + x));
		}
		for (; x < 4; ++x) {
			s += "   ";
		}
		System.out.println(s + "  " + i);
		if (System.out.checkError()) {
			System.exit(1);
		}
		return n;
	}

	public void dumpLine(int adr) {
		String str;
		str = String.format("%05x:", adr);
		int x;
		for (x = 0; x < 16; ++x) {
			if (adr + x < org || adr + x >= end) {
				str += "   ";
			} else {
				int c;
				c = read(adr + x);
				str += String.format(" %02x", c);
			}
		}
		str += "  ";
		for (x = 0; x < 16 && adr + x < end; ++x) {
			if (adr + x < org) {
				str += ' ';
			} else {
				int c;
				c = read(adr + x);
				if (c < ' ' || c > '~') {
					c = '.';
				}
				str += String.format("%c", (char)c);
			}
		}
		System.out.println(str);
		if (System.out.checkError()) {
			System.exit(1);
		}
	}
}
