import java.util.Arrays;
import java.util.Date;
import java.util.TimeZone;
import java.io.*;

// MP/M directory initialization for timestamps.
// No SFCBs, just LABEL with "Make XFCBs" set.

public class MpmInitDir implements CpmInitDir {
	private int user;
	private String filename;

	private byte[] dir;
	private long dirOff;
	private RandomAccessFile fd = null;
	private CpmDpb dpb;
	private InputStream data;
	private int[] sectrn = null;

	private int nextDirEnt; // index, not disk offset
	private int nextBlock; // index, not disk offset
	private int maxDirEnt;
	private int maxBlock;
	private boolean smallAlloc;
	private int blkSize;
	private int blkRecs;
	private int curRC;
	private int curExt;
	private int curS2;
	private int curEnt;
	private int curDirEnt;
	private int curDirOff;
	private int curBlock;
	private Date now = new Date();
	private TimeZone tz = TimeZone.getDefault();

	public MpmInitDir(File disk) {
		try {
			fd = new RandomAccessFile(disk, "rw");
		} catch (Exception ee) {
			ee.printStackTrace();
			return;
		}
	}

	public boolean setDpb(CpmDpb dpb) {
		dir = new byte[(dpb.drm + 1) * 32];
		dirOff = (dpb.off * dpb.spt + dpb.poff) * 128; // bytes
		try {
			fd.seek(dirOff);
			int n = fd.read(dir);
			if (n < 0) {
				System.err.format("EOF reading directory\n");
				return false;
			}
		} catch (Exception ee) {
			ee.printStackTrace();
			return false;
		}
		// Disk must be blank, initially.
		// It should have been just formatted, so not a problem.
		for (int x = 0; x <= dpb.drm; ++x) {
			int d0 = dir[x * 32] & 0xff;
			if (d0 != 0xe5) {
				System.err.format("Directory not empty (%d, %d, %d)\n", x, dirOff, dir.length);
				return false;
			}
		}
		int b = Integer.numberOfTrailingZeros(dpb.al0);
		if (b > 15) {
			System.err.format("DPB ALV0 not valid\n");
			return false;
		}
		b = 16 - b;
		this.dpb = dpb;
		nextDirEnt = 0;
		maxDirEnt = dpb.drm;
		if (dpb.plat > 1) {
			buildSecTab();
		}
		return true;
	}

	private void buildSecTab() {
		// operate translation in CP/M records...
		int pspl = (dpb.pssz / 128);
		sectrn = new int[dpb.spt];
		boolean[] secflg = new boolean[dpb.spt];
		Arrays.fill(sectrn, -1);
		Arrays.fill(secflg, false);
		int ntran = 0;
		int s = 0;
		while (ntran < dpb.spt) {
			while (secflg[s]) {
				s += pspl;
				if (s >= dpb.spt) s -= dpb.spt;
			}
			for (int x = 0; x < pspl; ++x) {
				secflg[s] = true;
				sectrn[ntran++] = s++;
			}
			s += (dpb.plat - 1) * pspl;
			if (s >= dpb.spt) s -= dpb.spt;
		}
//		String str = "sectrn[] = {";
//		for (int x = 0; x < dpb.spt; ++x) {
//			str += String.format(" %d", sectrn[x]);
//		}
//		System.err.format("%s }\n", str);
	}

	public boolean isError() { return fd == null; }

	public boolean close() {
		if (fd != null) {
			try {
				fd.close();
			} catch (Exception ee) {
				return false;
			}
		}
		return true;
	}

	private boolean closeDir() {
		if (fd != null) {
			if (sectrn == null) {
				try {
					fd.seek(dirOff);
					fd.write(dir);
				} catch (Exception ee) {
					return false;
				}
			} else {
				int dblks = (((dpb.drm + 1) / 4) + dpb.blm) >> dpb.bsh;
				int off = 0;
				for (int blk = 0; blk < dblks; ++blk) {
					writeBlock(blk, blkSize, dir, off);
					off += (128 << dpb.bsh);
				}
			}
		}
		return true;
	}

	// Must handle possible logical-to-physical sector translation...
	private boolean writeBlock(int blk, int num, byte[] buf, int boff) {
		if (sectrn == null) {
			long off = ((blk << dpb.bsh) * 128) + dirOff;
			try {
				fd.seek(off);
				fd.write(buf, boff, num);
			} catch (Exception ee) {
				return false;
			}
			return true;
		}
		// we know block 0 starts a track, but blocks may span 2 tracks...
		int recoff = (blk << dpb.bsh);
		for (int r = 0; r <= dpb.blm && num > 0; ++r) {
			int ps = (recoff + r) % dpb.spt;
			int toff = ((recoff + r) / dpb.spt) * dpb.spt;
			int ls = sectrn[ps];
			long off =  (toff + ls) * 128 + dirOff;
			try {
				fd.seek(off);
				fd.write(buf, boff, 128);
				boff += 128;
				num -= 128;
			} catch (Exception ee) {
				return false;
			}
		}
		return true;
	}

	private void unix2cpmdate(long time, byte[] buf, int start) {
		time += tz.getOffset(time);
		long unx = time / 1000;
		short date = (short)(unx / 86400 - 2922 + 1);
		unx %= 86400;
		int s = (int)(unx % 60); // seconds
		unx /= 60;
		int m = (int)(unx % 60); // minutes
		unx /= 60;
		int h = (int)(unx % 24); // hour
		buf[start++] = (byte)(date & 0xff);
		buf[start++] = (byte)((date >> 8) & 0xff);
		buf[start++] = (byte)(((h / 10) << 4) | (h % 10));
		buf[start++] = (byte)(((m / 10) << 4) | (m % 10));
		// No secs needed here...
	}

	private void setLabel(int p) {
		Arrays.fill(dir, 0, 32, (byte)0); // no timestamps?
		dir[0] = (byte)0x20;
		String lbl = String.format("PARTITN %d  ", p & 7);
		System.arraycopy(lbl.getBytes(), 0, dir, 1, 11);
		dir[12] = (byte)0b01110001; // Access, Update timestamps, Make XFCBs
		unix2cpmdate(now.getTime(), dir, 24);
		unix2cpmdate(now.getTime(), dir, 28);
	}

	public void doDirInit(int p, CpmDpb dpb) {
		if (!setDpb(dpb)) {
			System.err.format("Bad DPB, Skipping InitDir on %d\n", p);
			return;
		}
		// No SFCBs, let OS add XFCBs
		setLabel(p);
		closeDir();
		System.out.format("Initialized MP/M directory on partition %d\n", p);
	}
}
