import java.util.Arrays;
import java.io.*;

public class TD0DiskImage implements DiskImage {
	private byte[] img = null;
	private int[] sectbl;
	private int nsec;
	private int sec0;
	private int nbyt;
	private int lat;
	private int ntrk;
	private int nsid;
	private boolean density;
	private boolean interlaced;
	private byte[] sb = null;
	String label;
	int media; // 5 or 8
	int trkStart;
	int version;

	public TD0DiskImage(File f, boolean intl) {
		byte[] td0;
		try {
			td0 = new byte[(int)f.length()];
			FileInputStream fi = new FileInputStream(f);
			fi.read(td0);
			fi.close();
			if (td0.length < 4 || Character.toUpperCase(td0[0]) != 'T' ||
					Character.toUpperCase(td0[1]) != 'D' ||
					td0[2] != 0) {
				return;
			}
		} catch (Exception ee) {
			return;
		}
		nsec = 0;
		if (!check(td0)) {
			return;
		}
		img = td0;
		interlaced = intl;
		// override media size... seems to be wrong sometimes...
		if (ntrk == 77) {
			media = 8;
		} else if (ntrk == 40 || ntrk == 80) {
			media = 5;
		} else {
			return;
		}
	}

	public boolean isError() {
		return img == null;
	}

	public int length() {
		return img == null ? 0 : img.length;
	}

	class TD0Track {
		static final int sizeof = 4;
		public int spt;
		public int ptrk;
		public int psid;
		public int crc;

		public TD0Track(byte[] buf, int start) {
			spt = buf[start++] & 0xff;
			ptrk = buf[start++] & 0xff;
			psid = buf[start++] & 0xff;
			crc = buf[start++] & 0xff;
		}
	}

	class TD0Sector {
		static final int sizeof = 8;
		public int trk;
		public int sid;
		public int sec;
		public int sln;
		public int syn;
		public int crc;
		public int len;

		public TD0Sector(byte[] buf, int start) {
			trk = buf[start++] & 0xff;
			sid = buf[start++] & 0xff;
			sec = buf[start++] & 0xff;
			sln = buf[start++] & 0xff;
			syn = buf[start++] & 0xff;
			crc = buf[start++] & 0xff;
			len = buf[start++] & 0xff;
			len |= (buf[start++] & 0xff) << 8;
		}
	}

	class TD0Data20 {
		static final int sizeof = 1; // must add in length...
		public int len;
		public int _offset;

		public TD0Data20(byte[] buf, int start) {
			len = buf[start++] & 0xff;
			_offset = start;
		}
	}

	class TD0Data21 {
		static final int sizeof = 3;
		public int len_2; // length / 2, i.e. num "short"
		public int fil;

		public TD0Data21(byte[] buf, int start) {
			len_2 = buf[start++] & 0xff;
			fil = buf[start++] & 0xff;
			fil |= (buf[start++] & 0xff) << 8;
		}
	}

	class TD0Data2 {
		static final int sizeof = 1; // must add in length...
		public int typ;	// 0 or 1
		public int _offset; // start of TD0Data21 or TD0Data20

		public TD0Data2(byte[] buf, int start) {
			typ = buf[start++] & 0xff;
			_offset = start;
		}
	}

	class TD0Data1 {
		static final int sizeof = 4;
		public int len_2;	// num "short"
		public int fil;

		public TD0Data1(byte[] buf, int start) {
			len_2 = buf[start++] & 0xff;
			len_2 |= (buf[start++] & 0xff) << 8;
			fil = buf[start++] & 0xff;
			fil |= (buf[start++] & 0xff) << 8;
		}
	}

	class TD0Data {
		static final int sizeof = 1; // must add in d0 maybe...
		public int typ;
		public int _offset; // offset of "typ 0", TD0Data1 or TD0Data2

		public TD0Data(byte[] buf, int start) {
			typ = buf[start++] & 0xff;
			_offset = start;
		}
	}

	private void fillShorts(int fill, int num) {
		if (sb == null) {
			sb = new byte[nbyt];
		}
		for (int x = 0; x < nbyt && num > 0; ) {
			sb[x++] = (byte)(fill & 0xff);
			sb[x++] = (byte)((fill >> 8) & 0xff);
			--num;
		}
	}

	private int getTrack(byte[] b, int start, RandomAccessFile f) {
		int x = start;
		TD0Track tk;
		try {
			tk = new TD0Track(b, x);
		} catch (Exception ee) {
			ee.printStackTrace();
			return 0;
		}
		x += TD0Track.sizeof;
		if (tk.spt == 255) { // EOF
			return 0;
		}
		TD0Sector sc;
		try {
			sc = new TD0Sector(b, x);
		} catch (Exception ee) {
			ee.printStackTrace();
			return 0;
		}
		// What is this garbage in certain images?
		if (tk.spt == 1) {
			// will there ever be anything following?
			System.err.format("Nonsense track - assuming end of data\n");
			return 0;
		}
		int spt = tk.spt;
		int ssz = (128 << (sc.sln & 0x03));
		int trk = sc.trk;
		int sid = sc.sid;
		int sec = sc.sec;
		if (sc.sec >= spt) {
			sid = 1;
			sec -= spt;
		}
		int min = -1;
		if (nsec == 0) {
			nsec = spt;
			nbyt = ssz;
			lat = 1;
			sectbl = new int[nsec];
			min = spt;
		} else {
			if (nsec != spt || nbyt != ssz) {
				System.err.format("SPT/SSZ mismatch in TD0 file trk %d (%d,%d):(%d,%d)\n",
					trk, spt, ssz, nsec, nbyt);
System.err.format("dump %d: track %d %d %d\n", start, tk.spt, tk.ptrk, tk.psid);
System.err.format("dump %d: sector %d %d %d %d %d\n", x, sc.trk, sc.sid, sc.sec, sc.sln, sc.len);
				return -1;
			}
		}
		if (f == null) {
			if (ntrk > 0 && trk != ntrk) {
				// Too noisy for interlaced disks
				//System.err.format("Warning: track %d out of sequence\n", trk);
			}
			if (trk >= ntrk) ntrk = trk + 1;
			if (sid >= nsid) nsid = sid + 1;
		}
		int trkoff = 0;
		if (interlaced) {
			trkoff = (trk * nsid + sid) * spt * ssz;
		} else {
			trkoff = (sid * ntrk + trk) * spt * ssz;
		}
		for (int z = 0; z < spt; ++z) {
			sc = new TD0Sector(b, x);
			x += TD0Sector.sizeof;
			if (f == null && min >= 0) {
				// TODO: can first track ever be side 1?
				sectbl[z] = sc.sec;
				if (z == 1) {
					lat = sectbl[1] - sectbl[0];
				}
				if (sc.sec < min) {
					min = sc.sec;
				}
				x += sc.len;
				continue;
			}
			sec = sectbl[z];
			int off = trkoff + (sec - sec0) * ssz;
			TD0Data td = new TD0Data(b, x);
			switch(td.typ) {
			case 2:
				int y = x;
				x += TD0Data.sizeof;
				while (x - y < sc.len) {
					TD0Data2 td2 = new TD0Data2(b, x);
					switch(td2.typ) {
					case 0:
						TD0Data20 t20 = new TD0Data20(b, td2._offset);
						if (f != null) try {
							f.seek(off);
							f.write(b, t20._offset, t20.len);
						} catch (Exception ee) {
ee.printStackTrace();
							return -1;
						}
						off += t20.len;
						x += t20.len + TD0Data2.sizeof + TD0Data20.sizeof;
						break;
					case 1:
						TD0Data21 t21 = new TD0Data21(b, td2._offset);
						if (f != null) try {
							fillShorts(t21.fil, t21.len_2);
							f.seek(off);
							f.write(sb, 0, t21.len_2 * 2);
						} catch (Exception ee) {
ee.printStackTrace();
							return -1;
						}
						off += t21.len_2 * 2;
						x += 4;
						break;
					default:
						System.err.format("Format error type 2: %02x @ %d\n",
							td2.typ, x);
						return -1;
					}
				}
				continue;
			case 0:
				if (f != null) try {
					f.seek(off);
					f.write(b, td._offset, ssz);
				} catch (Exception ee) {
ee.printStackTrace();
					return -1;
				}
				off += ssz;
				x += sc.len;
				break;
			case 1:
				// fill sector td.d1.fil td.d1.len_2 shorts
				TD0Data1 td1 = new TD0Data1(b, td._offset);
				if (f != null) try {
					fillShorts(td1.fil, td1.len_2);
					f.seek(off);
					f.write(sb, 0, td1.len_2 * 2);
				} catch (Exception ee) {
ee.printStackTrace();
					return -1;
				}
				off += td1.len_2 * 2;
				x += sc.len;
				break;
			default:
				System.err.format("Format error type 0: %02x @ %d\n",
					td.typ, x - start);
				return -1;
			}
		}
		if (f == null && min >= 0) {
			sec0 = min;
		}
		return x - start;
	}

	class TD0Header {
		static final int sizeof = 12;
		public int vers;
		public int dens;
		public int dtype;
		public int tdens;
		public int dos;
		public int surf;
		public int crc;

		public TD0Header(byte[] buf, int start) {
			start += 4;
			vers = buf[start++] & 0xff;
			dens = buf[start++] & 0xff;
			dtype = buf[start++] & 0xff;
			tdens = buf[start++] & 0xff;
			dos = buf[start++] & 0xff;
			surf = buf[start++] & 0xff;
			crc = buf[start++] & 0xff;
			crc |= (buf[start++] & 0xff) << 8;
		}
	}

	class TD0Header2 {
		static final int sizeof = 10;
		int crc;
		int len;
		int yr;
		int mo;
		int dy;
		int hr;
		int mn;
		int se;

		public TD0Header2(byte[] buf, int start) {
			crc = buf[start++] & 0xff;
			crc |= (buf[start++] & 0xff) << 8;
			len = buf[start++] & 0xff;
			len |= (buf[start++] & 0xff) << 8;
			yr = buf[start++] & 0xff;
			mo = buf[start++] & 0xff;
			dy = buf[start++] & 0xff;
			hr = buf[start++] & 0xff;
			mn = buf[start++] & 0xff;
			se = buf[start++] & 0xff;
		}
	}

	private boolean check(byte[] td0) {
		if (td0[0] == 't' && td0[1] == 'd') {
			System.err.format("Advance compression TD0 not supported\n");
			return false;
		}
		int x = 0;
		TD0Header h1 = new TD0Header(td0, x);
		x += TD0Header.sizeof;
		//System.err.format("TD0 version %02x\n", h1.vers);
		if (h1.vers >= 0x15) {
			TD0Header2 h2 = new TD0Header2(td0, x);
			x += TD0Header2.sizeof;
			int z = h2.len;
			while (z > 0 && td0[x + z - 1] == 0) {
				--z;
			}
			label = String.format("%02d/%02d/%04d %02d:%02d:%02d\n",
				h2.mo, h2.dy, h2.yr + 1900, h2.hr, h2.mn, h2.se);
			label += new String(td0, x, z);
			x += h2.len;
		} else {
			label = "unlabelled";
		}
		version = h1.vers;
		media = h1.dtype == 2 ? 5 : 8; // a guess
		density = (h1.tdens & 1) != 0;
		trkStart = x; // second pass (write file) can start here...
		do {
			int n = getTrack(td0, x, null);
			if (n < 0) {
				return false;
			}
			if (n == 0) {
				break;
			}
			x += n;
		} while (x < td0.length);
		return true;
	}

	public boolean overlay(File f) {
		if (img == null) {
			return false;
		}
		RandomAccessFile fi = null;
		try {
			fi = new RandomAccessFile(f, "rw");
		} catch (Exception ee) {
			return false;
		}
		int x = trkStart;
		do {
			int n = getTrack(img, x, fi);
			if (n < 0) {
				return false;
			}
			if (n == 0) {
				break;
			}
			x += n;
		} while (x < img.length);
		return true;
	}

	public int mediaSize() { return media; }
	public int numSides() { return nsid; }
	public int numTracks() { return ntrk; }
	public int sectorsPerTrack() { return nsec; }
	public int sectorSize() { return nbyt; }
	public boolean interlaced() { return interlaced; }
	public int capacity() { return nsid * ntrk * nsec * nbyt; }
	public String getTitle() { return label; }
}
