// Copyright (c) 2020 Douglas Miller <durgadas311@gmail.com>
import java.util.Arrays;
import java.util.Properties;
import java.io.*;

// Protocol:
//	send: ENQ
//	recv: ACK
//	send: SOH <FMT> <DID> <SID> <FNC> <SIZ> <cks>
//	recv: ACK
//	send: STX <dat>... ETX <cks> EOT
//	recv: ACK
// cks = 0 - sum(SOH,...SIZ)
// cks = 0 - sum(STX,...)
// on recv, sum(SOH,...cks) == 0
//          sum(STX,...cks) == 0
// ASCII: bytes <FMT>...<cks> and <dat>...,<cks> are 2-digit hex values
// ASCII: <cks> computed on binary, not hex, values
//
// Any failure in the send will result in a retry, up to 10 times (total).
// The receive (must) also retry, for the matching number of times.
// This can be problematic if the sender and receiver get out of sync,
// for example when receive character timeouts occur.
// Note that the server needs to have a slightly different
// perspective on retries, since the client is expecting a message
// only in response to a request.
//
// 'debug' levels, cumulative
// 0 = no debug
// 1 = dump messages (success)
// 5 = dump transport/protocol errors
// 9 = dump all characters received

public class CpnetDRIProtocol implements CpnetSerialProtocol {
	static final int ENQ = 0x05;
	static final int SOH = 0x01;
	static final int STX = 0x02;
	static final int ETX = 0x03;
	static final int EOT = 0x04;
	static final int ACK = 0x06;
	static final int NAK = 0x15;
	int debug;
	int ctx;	// context for recv timeouts
	boolean ascii = false;
	InputStream in;
	OutputStream out;
	int cks = 0;

	// TODO: make these configurable
	// Timeouts must allow for interrupts on client.
	int charTimeout = 15;	// mS to wait for "next" character
	int ackTimeout = 15;	// mS to wait for ACK/NAK character
	int retries = 10;	// max number of attempts, total

	public CpnetDRIProtocol(Properties props, int dbg, String[] args) {
		debug = dbg;
		ascii = false;
		for (String ss : args) {
			if (ss.equalsIgnoreCase("ASCII")) {
				ascii = true;
			}
		}
		String s = props.getProperty("dri_ack_timeout");
		if (s != null) {
			ackTimeout = Integer.valueOf(s);
		}
		s = props.getProperty("dri_char_timeout");
		if (s != null) {
			charTimeout = Integer.valueOf(s);
		}
		s = props.getProperty("dri_retries");
		if (s != null) {
			retries = Integer.valueOf(s);
		}
	}

	public void setIO(InputStream in, OutputStream out) {
		this.in = in;
		this.out = out;
	}

	public int getRetries() { return retries; }

	private void chrOut(int c) throws Exception {
		out.write(c);
		out.flush();
	}

	private void ack() throws Exception {
		chrOut(ACK);
	}

	private void nak() throws Exception {
		chrOut(NAK);
	}

	// Returns -1 on comm failure (permanent)
	//         char on success (could be 00)
	private int chrIn() {
		try {
			int c = in.read();
			if (debug > 8) {
				System.err.format("> %02x\n", c);
			}
			return c;
		} catch (Exception ee) {
			ee.printStackTrace();
			return -1;
		}
	}

	// Returns -1 on comm failure (permanent)
	//         -2 timout
	//         char on success (could be 00)
	private int chrIn(int ms) {
		int tms = ms;
		try {
			while (in.available() == 0) {
				if (tms <= 0) {
					if (debug > 4) {
						System.err.format(
							"%d: tty recv timeout %d mS\n",
								ctx, ms);
					}
					return -2;
				}
				Thread.sleep(1);
				--tms;
			}
			int c = in.read();
			if (debug > 8) {
				System.err.format("> %02x\n", c);
			}
			return c;
		} catch (Exception ee) {
			ee.printStackTrace();
			return -1;
		}
	}

	// Returns -1 on comm failure (permanent)
	//         -2 timout
	//         char on success
	// Updates 'cks'
	private int netIn() throws Exception {
		int c = -1;
		c = chrIn(charTimeout);
		if (c < 0) return c;
		if (ascii) {
			int c1 = chrIn(charTimeout);
			if (c1 < 0) return c1;
			c = Character.digit(c, 16);
			c1 = Character.digit(c1, 16);
			c = (c << 4) | c1;
		}
		cks += c;
		return c;
	}

	// Returns 0 on success.
	//        -1 on permanent comm failure (do not retry)
	//         2 on timeout (client gone?)
	// checksum not received
	private int msgIn(byte[] buf, int off, int len) throws Exception {
		int c;
		for (int x = 0; x < len; ++x) {
			c = netIn();
			if (c == -1) return c;
			if (c < 0) return -c;
			int idx = off + x;
			if (idx < buf.length) {
				buf[idx] = (byte)c;
			}
		}
		return 0;
	}

	// Returns 0 on success.
	//        -1 on permanent comm failure (do not retry)
	//         1 on "NAK" (checksum error)
	//         2 on timeout (client gone?)
	//         3 on protocol error
	// Caller is responsible for reties.
	private int receive(byte[] buf, int off, int len) throws Exception {
		cks = SOH;
		ctx = 1;
		int c = msgIn(buf, off, NetworkServer.mhdrlen);
		if (c != 0) return c;
		ctx = 2;
		c = netIn();	// checksum byte
		if (c == -1) return c;
		if (c < 0) return -c;
		if ((cks & 0xff) == 0) {
			// success... keep going
			ack();
		} else {
			nak();
			if (debug > 4) {
				System.err.format("* SOH\n");
			}
			return 1;
		}
		ctx = 3;
		c = chrIn(charTimeout);
		if (c == -1) return c;
		if (c < 0) return -c;
		if (c != STX) {
			if (debug > 4) {
				System.err.format("! STX\n");
			}
			return 3;
		}
		if (debug > 8) {
			System.err.format("> STX\n");
		}
		cks = STX;
		int mlen = (buf[off + NetworkServer.SIZ] & 0xff) + 1;
		ctx = 4;
		c = msgIn(buf, off + NetworkServer.DAT, mlen);
		if (c != 0) return c;
		ctx = 5;
		c = chrIn(charTimeout);
		if (c == -1) return c;
		if (c < 0) return -c;
		if (c != ETX) {
			if (debug > 4) {
				System.err.format("! ETX\n");
			}
			return 3;
		}
		cks += c;
		ctx = 6;
		c = netIn();	// checksum byte, only need resulting 'cks'
		if (c == -1) return c;
		if (c < 0) return -c;
		ctx = 7;
		c = chrIn(charTimeout);
		if (c == -1) return c;
		if (c < 0) return -c;
		if (c != EOT) {
			if (debug > 4) {
				System.err.format("! EOT\n");
			}
			return 3;
		}
		if ((cks & 0xff) == 0) {
			// success...
			ack();
			return 0;
		} else {
			nak();
			if (debug > 4) {
				System.err.format("* STX\n");
			}
			return 1;
		}
	}

	public boolean protoStart(int ch) {
		return (ch == ENQ);
	}

	// Returns 0 on success.
	//        -1 on permanent comm failure (do not retry)
	//         1 on "NAK" (checksum error)
	//         2 on timeout (client gone?)
	//         3 on protocol error
	// Caller is responsible for retries
	public int recv(byte[] buf, int off, int len) {
		int mlen = 0;
		try {
			while (true) {
				int c = chrIn(); // might as well block here...
				if (c < 0) return -1;
				if (c == ENQ) {
					ack();
					if (debug > 8) {
						System.err.format("> ENQ\n");
					}
				} else if (c == SOH) {
					if (debug > 8) {
						System.err.format("> SOH\n");
					}
					c = receive(buf, off, len);
					return c;
				} else {
					// error? keep trying?
					return 3;
				}
			}
		} catch (Exception ee) {
			ee.printStackTrace();
			return -1;
		}
	}

	// Same as recv(byte[] buf, int off, int len) except already have
	// first character.
	public int recv(int ch, byte[] buf, int off, int len) {
		if (ch != ENQ) { // major program error...
			return 3;
		}
		if (debug > 8) {
			System.err.format("> ENQ\n");
		}
		int mlen = 0;
		try {
			ack();
			while (true) {
				ctx = 0;
				int c = chrIn(); // might as well block here...
				if (c < 0) return -1;
				if (c == ENQ) {
					ack();
					if (debug > 8) {
						System.err.format("> ENQ\n");
					}
				} else if (c == SOH) {
					if (debug > 8) {
						System.err.format("> SOH\n");
					}
					ctx = 2;
					c = receive(buf, off, len);
					return c;
				} else {
					// error? keep trying?
					return 3;
				}
			}
		} catch (Exception ee) {
			ee.printStackTrace();
			return -1;
		}
	}

	// Returns 0 on success.
	//        -1 on permanent comm failure (do not retry)
	//         1 on "NAK" (at least, not ACK)
	//         2 on timeout (client gone?)
	private int getACK() throws Exception {
		int c = chrIn(ackTimeout);
		if (c == ACK) return 0;
		if (c == -1) return c;
		if (c < 0) return -c;
		return 1;
	}

	private void nibble(int val) throws Exception {
		out.write(Character.toUpperCase(Character.forDigit(val, 16)));
	}

	private void netOut(int c) throws Exception {
		if (ascii) {
			nibble(c >> 4);
			nibble(c & 0x0f);
		} else {
			out.write(c);
		}
		out.flush();
		cks += c;
	}

	private void msgOut(byte[] buf, int off, int len) throws Exception {
		int d;
		for (int x = 0; x < len; ++x) {
			d = buf[off + x] & 0xff;
			netOut(d);
		}
	}

	// Returns 0 on success.
	//        -1 on permanent comm failure (do not retry)
	//         1 on "NAK"
	//         2 on timeout (client gone?)
	// Caller is responsible for reties.
	public int send(byte[] buf, int off, int len) {
		try {
			ctx = 10;
			// Note: client should be waiting for us.
			chrOut(ENQ);
			int c = getACK();
			if (c != 0) return c;
			++ctx;	// ctx = 11
			chrOut(SOH);
			cks = SOH;
			msgOut(buf, off, NetworkServer.mhdrlen);
			netOut((0 - cks) & 0xff);
			c = getACK();
			if (c != 0) return c;
			++ctx;	// ctx = 12
			len -= NetworkServer.mhdrlen;
			chrOut(STX);
			cks = STX;
			msgOut(buf, off + NetworkServer.mhdrlen, len);
			chrOut(ETX);
			cks += ETX;
			netOut((0 - cks) & 0xff);
			chrOut(EOT);
			c = getACK();
			++ctx;	// ctx = 13
			return c;
		} catch (Exception ee) {
			ee.printStackTrace();
			return -1;
		}
	}
}
