// Copyright (c) 2020 Douglas Miller <durgadas311@gmail.com>
import java.io.*;
import java.util.Properties;

public class CpnetSimpleProtocol implements CpnetSerialProtocol {
	int debug;
	boolean ascii = true;
	boolean usecrc = true;
	InputStream in;
	OutputStream out;

	public CpnetSimpleProtocol(Properties props, int dbg, String[] args) {
		if (args == null) return;
		debug = dbg;
		ascii = false;
		usecrc = false;
		for (String ss : args) {
			if (ss.equalsIgnoreCase("ASCII")) {
				ascii = true;
			} else if (ss.equalsIgnoreCase("CRC")) {
				usecrc = true;
			}
		}
	}

	public void setIO(InputStream in, OutputStream out) {
		this.in = in;
		this.out = out;
	}

	public int getRetries() { return 0; }

	public boolean protoStart(int ch) {
		return false; // This proto does not support console mode
	}

	public int recv(byte[] buf, int off, int len) {
		int n = 0;
		int idx = -1;
		int mlen = 256;
		int v = 0x0f;
		while (true) {
			int c = -1;
			try {
				c = in.read();
				if (c < 0) {
					return -1;
				}
			} catch (Exception ee) {
				ee.printStackTrace();
				return -1;
			}
			if (idx < 0) {	// idle
				if (ascii) {
					if (c != '+') {
						// error, really
						idx = -1;
						continue;
					}
				} else {
					crc = 0xffff;
					idx = NetworkServer.FMT;
				}
			}
			if (ascii) {
				if (c == '+') {
					crc = 0xffff;
					idx = NetworkServer.FMT;
					v = 0x0f;
					continue;
				}
				if (c == '-') {	// EOM
					if (usecrc && crc != 0) {
						System.err.format("CRC error (%04x)\n",
									crc);
						// TODO: what to do? send NAK?
						return 1;
					}
					return  0; // "success"
				}
				int d = Character.digit(c, 16);
				if (d < 0) {
					// TODO: be more, or less, draconian?
					idx = -1;
					continue;
				}
				v <<= 4;
				v |= d;
				if (v <= 255) {
					continue;
				}
			} else {
				if (idx == NetworkServer.DAT) {
					mlen = (buf[NetworkServer.SIZ] & 0xff) + 1 +
							NetworkServer.mhdrlen;
				}
				v = c;
			}
			// TODO: return truncated message status...
			crcByte(v & 0xff);
			if (idx < buf.length) {
				buf[idx++] = (byte)v;
				++n;
				if (!ascii && n >= mlen) {
					return 0; // success
				}
			}
			v = 0x0f; // only for ascci, but nop else
		}
	}

	public int recv(int ch, byte[] buf, int off, int len) {
		return 3;	// not supported
	}

	private int crc;
	static final int POLY = 0x8408;
	private void crcByte(int data) {
		int i;
		int mask;

		for (i = 0; i < 8; ++i) {
			mask = ((crc ^ data) & 1);
			crc >>= 1;
			data >>= 1;
			if (mask != 0) {
				crc ^= POLY;
			}
		}
	}

	private void nibble(int val) throws Exception {
		out.write(Character.toUpperCase(Character.forDigit(val, 16)));
	}

	// Returns 0 on success.
	// 	  -1 on permanent comm failure (do not retry)
	public int send(byte[] buf, int off, int len) {
		try {
			if (ascii) {
				out.write('+');
				out.write('+');
			}
			crc = 0xffff;
			int d;
			for (int x = 0; x < len; ++x) {
				d = buf[off + x] & 0xff;
				crcByte(d);
				if (ascii) {
					nibble(d >> 4);
					nibble(d & 0x0f);
				} else {
					out.write(d);
				}
			}
			if (ascii) {
				if (usecrc) {
					d = crc & 0xff;
					nibble(d >> 4);
					nibble(d & 0x0f);
					d = (crc >> 8) & 0xff;
					nibble(d >> 4);
					nibble(d & 0x0f);
				}
				out.write('-');
				out.write('-');
			} else {
				if (usecrc) {
					out.write(crc & 0xff);
					out.write((crc >> 8) & 0xff);
				}
			}
			out.flush();
			return 0;
		} catch (Exception ee) {
			ee.printStackTrace();
			return -1;
		}
	}
}
