// Copyright (c) 2020 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Properties;
import java.io.*;

public class EEPROM implements Rom, TimeListener, Runnable {
	protected String name;
	protected RandomAccessFile flash;
	protected java.util.concurrent.LinkedBlockingDeque<Boolean> wrq;
	protected byte[] rom;
	protected byte[] pag;
	protected int pgz;
	protected int pgm;
	protected int org;

	private int gpioIO1 = 0b10000000;
	private int lastPg = -1;
	private int lastWr = -1;
	private int lastDa = -1;
	private boolean pending;
	private boolean flashing;
	private boolean wEnabled;
	private boolean sdp;
	private int sdpState = 0;
	private Thread t;
	private long ntime = 0;

	public EEPROM(Properties props, String id, int base, Interruptor intr) {
		name = id;
		org = base;
		pgz = 64;	// TODO: variable page size
		pgm = pgz - 1;	// assumes power of 2
		String pfx = id.toLowerCase();
		wrq = new java.util.concurrent.LinkedBlockingDeque<Boolean>();
		rom = new byte[0x8000];
		pag = new byte[pgz];
		String s = props.getProperty(pfx + "_backingstore");
		if (s == null) {
			s = pfx + ".eeprom";
		}
		try {
			flash = new RandomAccessFile(s, "rw");
			if (flash.length() == 0) {
				Arrays.fill(rom, (byte)0xff);
				flash.write(rom);
			} else {
				// TODO: validate size?
				flash.read(rom);
			}
		} catch (Exception ee) {
			System.err.format("%s media \"%s\" fail: %s\n", id,
							s, ee.getMessage());
			return;
		}
		intr.addTimeListener(this);
		Thread t = new Thread(this);
		t.start();
	}

	public int length() { return 0x8000; }

	public int read(int addr) {
		addr &= 0x7fff;
		if (flashing) {
			//if (lastWr == addr) {	// TODO: yes? /DATA polling
			// TODO: how does toggle bit work?
			// lastDa ^= 0x40;	// toggle bit 6?
			return lastDa ^ 0x80;
		}
		return rom[addr] & 0xff;
	}

	private void chkSdp(int addr, int val) {
		switch (sdpState) {
		case 0:
			if (lastPg < 0 && addr == 0x5555 && val == 0xaa) {
				++sdpState;
				return;
			}
			break;
		case 1:	// saw 0x5555 0xaa
			if (addr == 0x2aaa && val == 0x55) {
				++sdpState;
				return;
			}
			break;
		case 2:	// saw 0x2aaa 0x55
			if (addr != 0x5555) break;
			if (val == 0xa0) {	// SDP on
				sdp = true;
				lastPg = -1;
				pending = true;
				wrq.add(false);
			} else if (val == 0x80) {
				++sdpState;
				return;
			}
			break;
		case 3:	// saw 0x5555 0x80	// SDP off, stage 1
			if (addr == 0x5555 && val == 0xaa) {
				++sdpState;
				return;
			}
			break;
		case 4:	// saw 0x5555 0xaa (2nd)
			if (addr == 0x2aaa && val == 0x55) {
				++sdpState;
				return;
			}
			break;
		case 5:	// saw 0x2aaa 0x55 (2nd)
			if (addr == 0x5555 && val == 0x20) {	// SDP off, final
				sdp = false;
				lastPg = -1;
				pending = true;
				wrq.add(false);
			}
			break;
		default:
			break;
		}
		sdpState = 0;
	}

	// caller ensures addr < length()
	public void write(int addr, int val) {
		if (!wEnabled || flashing) return;
		addr &= 0x7fff;
		if (addr == 0x5555 || addr == 0x2aaa) chkSdp(addr, val);
		int pga = addr & ~pgm;	// page address
		if (lastPg < 0) {	// idle?
			System.arraycopy(rom, pga, pag, 0, pgz);
			lastPg = pga;
			pending = true;
		} else if (lastPg != pga) {
			// TODO: ignore invalid addresses? restart?
			return;
		}
		pag[addr & pgm] = (byte)val;
		lastWr = addr;
		lastDa = val;
		wrq.add(sdp);	// might be ignored if SDP on/off preceded
	}

	public void writeEnable(boolean we) {
		// trigger write on trailing edge?
		if (wEnabled != we) {
			wEnabled = we;
		}
	}

	public void addTime(int nsec) {
		ntime += nsec;
	}

	public void run() {
		while (true) {
			boolean prot = true;
			try {
				prot = wrq.take();
				// Need to wait 150uS after last write,
				// but simulator backlog sleep perterbs this.
				// Need to time this to CPU cycles, not real-time.
				long t0 = ntime;
				while (ntime - t0 < 150000) {
					if (wrq.size() > 0) {
						prot = wrq.take();
						t0 = ntime;
					}
				}
				flashing = true;
				pending = false;
				if (!prot && lastPg >= 0) {
					System.arraycopy(pag, 0, rom, lastPg, pgz);
					if (flash != null) {
						// TODO: always write the whole thing?
						flash.seek(0);
						flash.write(rom);
					}
				}
				Thread.sleep(10);	// 10mS write time
			} catch (Exception ee) {
				System.err.format("%s write fail: %s\n",
							name, ee.getMessage());
			} finally {
				wrq.clear();
				flashing = false;
				lastPg = -1;
			}
		}
	}

	public String dumpDebug() {
		String str = String.format("%s EEPROM at %04x\n", name, org);
		// TODO: more info...
		str += String.format("sdp=%s sdpState=%d\n",
			sdp, sdpState);
		str += String.format("WE=%s wrq=%d flashing=%s pending=%s\n",
			wEnabled, wrq.size(),flashing, pending);
		str += String.format("lastPg=%04x lastWr=%04x lastDa=%02x\n",
			lastPg, lastWr, lastDa);
		return str;
	}
}
