// Copyright (c) 2016 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Properties;
import java.util.Vector;
import java.awt.event.*;
import javax.swing.Timer;
import java.io.*;
import java.lang.reflect.Constructor;

public class FT245RDevice
		implements IODevice, VirtualUART, ActionListener {

	static final int dataPortOffset = 0;
	static final int statusPortOffset = 2;
	static final int sts_DataReady = 0x02;
	static final int sts_TxReady = 0x01;

	private int basePort;
	private java.util.concurrent.LinkedBlockingDeque<Integer> fifo;
	private java.util.concurrent.LinkedBlockingDeque<Integer> fifi;

	private byte[] buf;
	private byte[] respBuf = null; // might point to 'buf'
	private int bufoff = 0;
	private int buflen = 0;
	private int last = 0;

	private Object attObj;
	private OutputStream attFile;
	private IODevice vdip1;

	private LED led;
	javax.swing.Timer timer;

	public FT245RDevice(Properties props, int base, IODevice vdip,
			LEDHandler lh, Interruptor intr) {
		super();
		attObj = null;
		attFile = null;
		basePort = base;
		vdip1 = vdip;
		fifo = new java.util.concurrent.LinkedBlockingDeque<Integer>();
		fifi = new java.util.concurrent.LinkedBlockingDeque<Integer>();
		led = lh.registerLED(getDeviceName());
		timer = new javax.swing.Timer(500, this);
		System.err.format("FT245R device at port %02x\n", basePort);
		String s = props.getProperty("ft245rdevice_att");
		if (s != null && s.length() > 1) {
			// The only supported attachment, for now...
			attachClass(props, s);
		}
	}

	private void setupFile(String[] args, int start) {
		boolean append = false;
		for (int x = start + 1; x < args.length; ++x) {
			if (args[x].equals("a")) {
				append = true;
			}
		}
		if (args[start].equalsIgnoreCase("syslog")) {
			attFile = System.err;
		} else try {
			attFile = new FileOutputStream(args[start], append);
		} catch (Exception ee) {
			System.err.format("Invalid file in attachment: %s\n", args[start]);
			return;
		}
	}

	private void attachClass(Properties props, String s) {
		String[] args = s.split("\\s");
		for (int x = 1; x < args.length; ++x) {
			if (args[x].startsWith(">")) {
				args[x] = args[x].substring(1);
				setupFile(args, x);
				// TODO: truncate args so Class doesn't see?
			}
		}
		Vector<String> argv = new Vector<String>(Arrays.asList(args));
		// try to construct from class...
		try {
			Class<?> clazz = Class.forName(args[0]);
			Constructor<?> ctor = clazz.getConstructor(
					Properties.class,
					argv.getClass(),
					VirtualUART.class);
			// funky "new" avoids "argument type mismatch"...
			attObj = ctor.newInstance(
					props,
					argv,
					(VirtualUART)this);
		} catch (Exception ee) {
			System.err.format("Invalid class in attachment: %s\n", s);
			return;
		}
	}

	public int getBaseAddress() { return basePort; }
	public int getNumPorts() { return 4; } // actually only 3, and one not used.
	public String getDeviceName() { return "FT245R"; }

	private boolean ledState = false;

	private void startLED() {
		if (!ledState && led != null) {
			led.set(true);
			ledState = true;
		}
		timer.removeActionListener(this);
		timer.addActionListener(this);
		timer.restart();
	}

	public void reset() {
		//super.reset();
		// clear registers...
		fifo.clear();
		fifi.clear();
	}

	public int in(int port) {
		int off = port - basePort;
		int val = 0;
		if (off == statusPortOffset) {
			if (fifi.size() > 0) {
				val |= sts_DataReady;
			}
			val |= sts_TxReady;
			if (vdip1 != null) {
				val |= vdip1.in(port);
			}
		} else if (off == dataPortOffset) {
			// must not sleep.
			if (fifi.size() > 0) {
				try {
					val = fifi.take();
// should all be printable ASCII...
//System.err.format("%c", (char)val);
				} catch (Exception ee) {}
			}
		} else if (vdip1 != null) {
			val = vdip1.in(port);
		}
		return val;
	}

	public void out(int port, int val) {
		int off = port - basePort;
		if (off == dataPortOffset) {
			startLED();
			if (attFile != null) try {
				attFile.write(val);
			} catch (Exception ee) {}
			if (attObj != null) {
				fifo.add(val & 0xff);
			}
		} else if (vdip1 != null) {
			vdip1.out(port, val);
		}
	}

	// VirtualUART, how the server dispatch accesses the data stream
	public boolean attach(Object periph) { return false; }
	public void detach() {
		if (attObj == null) {
			return;
		}
		attObj = null;
		try {
			fifo.addFirst(-1);
		} catch (Exception ee) {
			fifo.add(-1);
		}
	}
	public int available() {
		return fifo.size();
	}
	public int take() {
		try {
			int c = fifo.take();
			return c;
		} catch(Exception ee) {
			return -1;
		}
	}
	public boolean ready() {
		return true;
	}
	public void put(int ch, boolean sleep) {
		startLED();
		// TODO: prevent infinite growth?
		fifi.add(ch & 0xff);
	}
	public void setModem(int mdm) { }
	public int getModem() { return 0xff; }
	public void attachDevice(SerialDevice io) { }
	public String getPortId() { return "ft245rdevice"; }

	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == timer) {
			timer.removeActionListener(this);
			if (led != null) {
				led.set(false);
			}
			ledState = false;
		}
	}

	public String dumpDebug() {
		String ret = new String();
		ret += String.format("port %02x", basePort);
		ret += String.format("\n#fifo = %d", fifo.size());
		if (fifo.size() > 0) {
			ret += String.format(" %02x", fifo.peek());
		}
		ret += String.format("\n#fifi = %d", fifi.size());
		if (fifi.size() > 0) {
			ret += String.format(" %02x", fifi.peek());
		}
		ret += '\n';
		return ret;
	}

}
