// Copyright 2016 Douglas Miller

import java.util.Arrays;
import java.util.Vector;
import java.util.Map;
import java.util.HashMap;
import java.io.*;
import java.util.concurrent.Semaphore;
import java.util.concurrent.locks.ReentrantLock;
import java.util.Properties;

import z80core.*;

public class H89 implements Computer, H89Commander, Interruptor, GppListener, Runnable {
	private CPU cpu;
	private long clock;
	private Map<Integer, IODevice> ios;
	private Vector<IODevice> devs;
	private Vector<DiskController> dsks;
	private Vector<InterruptController> intrs;
	private Memory mem;
	private VirtualUART con;
	private GeneralPurposePort gpp;
	private H8FrontPanel fp;
	private boolean running;
	private boolean stopped;
	private Semaphore stopWait;
	private boolean ssen;	// single-step enabled (on)
	private int ssbits;
	private int sscnt = 0;
	private boolean tracing;
	private int traceCycles;
	private int traceLow;
	private int traceHigh;
	private int[] intRegistry;
	private int[] intLines;
	private int intState;
	private int intMask;
	private Vector<ClockListener> clks;
	private Vector<TimeListener> times;
	private	int cpuSpeed = 2048000;
	private int cpuCycle2ms = 4096;
	private int nanoSecCycle = 488;
	private Z80Disassembler disas;
	private ReentrantLock cpuLock;

	// Missed 2mS interrupt statistics
	private long backlogNs;

	// For now, try to make H8 a variation of an H89...
	private boolean isH8;

	public static boolean isH8(Properties props) {
		String s = props.getProperty("model");
		// H8, WH8, H8A, WH8A (...)
		return (s != null && s.matches("W?H8A?"));
	}

	public static boolean isNorby(Properties props) {
		String s = props.getProperty("h8_cpu");
		return (s != null && (s.startsWith("Z80_RTC_CLOCK_ORG0") ||
					s.startsWith("Z180_RTC_CLOCK_ORG0")));
	}

	public static boolean isNorbyFeat(Properties props, String feat) {
		String s = props.getProperty("h8_cpu");
		return (s != null && s.indexOf("_" + feat) > 0);
	}

	public static boolean isDuart(Properties props) {
		return isNorbyFeat(props, "DUART");
	}

	public static boolean isZ80(Properties props) {
		if (!isH8(props)) {
			return true;
		}
		String s = props.getProperty("h8_cpu");
		return (s != null && s.indexOf("Z80") >= 0);
	}

	public static boolean isZ180(Properties props) {
		if (!isH8(props)) {
			return false;
		}
		String s = props.getProperty("h8_cpu");
		return (s != null && s.indexOf("Z180") >= 0);
	}

	public H89(Properties props, LEDHandler lh) {
		String s;
		intRegistry = new int[8];
		intLines = new int[8];
		Arrays.fill(intRegistry, 0);
		Arrays.fill(intLines, 0);
		intState = 0;
		intMask = 0;
		running = false;
		stopped = true;
		stopWait = new Semaphore(0);
		cpuLock = new ReentrantLock();
		boolean is_z180 = isZ180(props);
		Z180 z180 = null;
		if (is_z180) {
			z180 = new Z180(this);
			cpu = z180;
			//setSpeed(16384000); // what is speed?
			// ... add Z180 devices... listeners?
			props.setProperty("h89ena2", "1");
		} else if (isZ80(props)) {
			cpu = new Z80(this);
		} else {
			cpu = new I8080(this);
		}
		ios = new HashMap<Integer, IODevice>();
		devs = new Vector<IODevice>();
		dsks = new Vector<DiskController>();
		clks = new Vector<ClockListener>();
		times = new Vector<TimeListener>();
		intrs = new Vector<InterruptController>();
		// Do this early, so we can log messages appropriately.
		s = props.getProperty("h89_log");
		if (s != null) {
			String[] args = s.split("\\s");
			boolean append = false;
			for (int x = 1; x < args.length; ++x) {
				if (args[x].equals("a")) {
					append = true;
				}
			}
			try {
				FileOutputStream fo = new FileOutputStream(args[0], append);
				PrintStream ps = new PrintStream(fo, true);
				System.setErr(ps);
			} catch (Exception ee) {}
		}
		s = props.getProperty("configuration");
		if (s == null) {
			System.err.format("No config file found\n");
		} else {
			System.err.format("Using configuration from %s\n", s);
		}
		s = props.getProperty("h89_trace");
		if (s != null) {
			Vector<String> ret = new Vector<String>();
			traceCommand(s.split("\\s"), ret, ret);
			// TODO: log error?
		}

		boolean has_serial = false;
		isH8 = H89.isH8(props);
		if (isH8) {
			has_serial = configH8(props, lh);
		} else {
			has_serial = configH89(props, lh);
		}
		ssbits = gpp.getSStepBit();
		gpp.addGppListener(this);

		if (false && has_serial) {
			for (int p = 0xd0; p < 0xe8; p += 8) {
				IODevice dev = getDevice(p);
				if (dev instanceof INS8250) {
					INS8250 sp = (INS8250)dev;
					sp.setModem(VirtualUART.SET_CTS | VirtualUART.SET_DSR);
				}
			}
		}
		s = props.getProperty("h89_disas");
		if (s != null && s.equalsIgnoreCase("zilog")) {
			if (is_z180) {
				disas = new Z180DisassemblerZilog(mem, z180);
			} else {
				disas = new Z80DisassemblerZilog(mem);
			}
		} else {
			if (is_z180) {
				disas = new Z180DisassemblerMAC80(mem, z180);
			} else {
				disas = new Z80DisassemblerMAC80(mem);
			}
		}
	}

	private boolean configH8(Properties props, LEDHandler lh) {
		String s;
		String[] slots = new String[] {"slot_p1", "slot_p2", "slot_p3",
			"slot_p4", "slot_p5", "slot_p6",
			"slot_p7", "slot_p8", "slot_p9", "slot_p10"
			};
		mem = null;
		int h8min = 16;
		int h8max = 0;
		IODevice cpn = null;
		IODevice rtc = null;
		boolean has_serial = false;
		boolean noMoreMem = false;
		if (isNorby(props)) {
			// Norberto's new Z80 or Z180 board, with RAM, RTC, DUART.
			NorbysGpp h8gpp = new NorbysGpp(props, this);
			// Only Norberto's CPU boards can run this way, a la H89
			s = props.getProperty("h8nofp");
			if (s == null) {
				fp = new H8FrontPanel(props, "h8fp", this, h8gpp, this);
				// TODO: else auto-enable H89 NMI?
			}
			s = props.getProperty("h8nmi");
			if (s != null) {
				addDevice(new LegacyH8Ports(props, 0xf0, 2, this));
				addDevice(new LegacyH8Ports(props, 0xfa, 2, this));
			}
			gpp = h8gpp;
			addDevice(gpp);

			if (isZ180(props)) {
				noMoreMem = true;
				mem = new Z180Memory(props, gpp, this);
			} else {
				mem = new MMS77311Memory(props, gpp, this);
			}
			// NOTE: these are NOT H8-4 ports!
			boolean duart = isDuart(props);
			s = props.getProperty("jp10");
			if (s == null || s.equals("1")) {
				if (duart) {
					props.setProperty("con_chip", "TL16C2550");
				}
				INS8250 sp = new INS8250(props, false, "con", 0xe8, 3, this);
				con = sp;
				addDevice(sp);
				con.setModem(VirtualUART.SET_CTS | VirtualUART.SET_DSR);
				has_serial = true;
			}
			s = props.getProperty("jp4");
			if (s != null && s.equals("1")) {
				if (duart) {
					props.setProperty("aux_chip", "TL16C2550");
				}
				addDevice(new INS8250(props, false, "aux", 0xd0, 5, this));
				has_serial = true;
			} else if (s == null || s.equals("3")) {
				if (duart) {
					props.setProperty("lp_chip", "TL16C2550");
				}
				addDevice(new INS8250(props, false, "lp", 0xe0, 5, this));
				has_serial = true;
			}
			s = props.getProperty("jp14");
			if (s == null || s.equals("1")) {
				// TODO: make this device standalone... w/o FT245RDevice
				// or, make FT245RDevice "invisible"...
				IODevice vdip = new VDIP1Device(props, 0xd8, lh, this);
				addDevice(vdip);
				// FT245R must exist here?
				cpn = new FT245RDevice(props, 0xd8, vdip, lh, this);
			}
			// TODO: need to allow partial config
			if (isNorbyFeat(props, "GIDE")) {
				addDevice(new GIDE_CF(props, null, 0x80, lh, this));
			}
			rtc = new RTC72421(props, 0xa0);
			addDevice(rtc);
		} else {
			H8Gpp h8gpp = new H8Gpp(props, this);
			fp = new H8FrontPanel(props, "h8fp", this, h8gpp, this);
			gpp = h8gpp;
			addDevice(gpp);
		}
		for (String slot : slots) {
			s = props.getProperty(slot);
			if (s == null) {
				continue;
			}
			if (s.matches("W?H8-[13].*")) {
				if (mem != null) {
					System.err.format("WARNING: %s.%s: ignoring redundant memory\n", slot, s);
					continue;
				}
				int lo = 2;
				int sz = 1;
				if (s.charAt(3) == '3' || s.charAt(4) == '3') {
					sz = 2;
				}
				String[] ss = s.split("\\s");
				if (ss.length > 1 && ss[1].matches("[0-9]*")) {
					lo = Integer.valueOf(ss[1]) / 8;
					if (lo > 7) lo = 7;
					lo *= 2;
				}
				if (lo < h8min) h8min = lo;
				if (lo + sz > h8max) h8max = lo + sz;
				continue;
			}
			if (s.matches("W?H8-64")) {
				if (mem != null) {
					System.err.format("WARNING: %s.%s: ignoring redundant memory\n", slot, s);
					continue;
				}
				mem = new MMS77311Memory(props, gpp, this);
				continue;
			}
			if (s.startsWith("X2H8") ||
					s.startsWith("X/2-H8")) {
				if (mem != null) {
					if (noMoreMem) {
						System.err.format("WARNING: %s.%s: ignoring redundant memory\n", slot, s);
						continue;
					} else {
						System.err.format("WARNING: %s.%s: replacing existing memory\n", slot, s);
					}
				}
				X2H8Memory xm = new X2H8Memory(props, 0x00, gpp, this);
				addDevice(xm);
				mem = xm;
				continue;
			}
			if (s.startsWith("H8-512K")) {
				// TODO: any error checking required?
				if (mem != null) {
					if (noMoreMem) {
						System.err.format("WARNING: %s.%s: ignoring redundant memory\n", slot, s);
						continue;
					} else {
						System.err.format("WARNING: %s.%s: replacing existing memory\n", slot, s);
					}
				}
				H8_512KMemory xm = new H8_512KMemory(props, 0x00, gpp, this);
				addDevice(xm);
				mem = xm;
				continue;
			}
			if (s.equals("H8xSPI_CF") || s.equals("H8xSPI-CF")) {
				addDevice(new H8xSPI(props, "H8xSPI", 0x40, 5, lh, this));
				addDevice(new GIDE_CF(props, "CF", 0x80, lh, this));
				continue;
			} else if (s.startsWith("H8xSPI")) {
				int port = 0x40;
				if (s.length() > 6 && s.charAt(6) == '2') {
					port = 0x42;
				}
				addDevice(new H8xSPI(props, s, port, 5, lh, this));
				continue;
			}
			if (s.equals("RTC72421") || s.equals("H8RTCETH")) {
				if (rtc != null) {
					System.err.format("WARNING: %s.%s: ignoring redundant RTC\n", slot, s);
					continue;
				}
				rtc = new RTC72421(props, 0xa0);
				addDevice(rtc);
				continue;
			}
			if (s.matches("W?H8-5")) {
				// TODO: any conflict with default H8-4?
				// This disables H8-4 "console" port...
				INS8251 sp;
				if (con != null) {
					System.err.format("WARNING: %s.%s: console already set, using CON2\n", slot, s);
					sp = new INS8251(props, "con2", 0xfa, 6, this);
				} else {
					sp = new INS8251(props, "con", 0xfa, 6, this);
					con = sp;
					con.setModem(VirtualUART.SET_CTS | VirtualUART.SET_DSR);
				}
				addDevice(sp);
				sp = new INS8251(props, "cass", 0xf8, 4, this);
				addDevice(sp);
				H8CassetteDrive cass = new H8CassetteDrive(props, lh, this);
				sp.attachDevice(cass);
				addDiskDevice(cass); // not really an IODevice...
				continue;
			}
			if (s.matches("W?H8-4")) {
				// TODO: support multiple cards? Partial configs?
				if (has_serial) {
					System.err.format("WARNING: %s.%s: ignoring duplicate serial ports\n", slot, s);
					continue;
				}
				// This is the default, handle outside loop...
				continue;
			}
			if (s.equals("H8GIDE")) {
				addDevice(new GIDE_CF(props, null, 0x80, lh, this));
				continue;
			}
			if (s.equals("H8USB")) {
				int port = 0xb0;
				s = props.getProperty("h8usb_port");
				if (s != null) try {
					port = Integer.decode(s);
				} catch (Exception ee) {}
				IODevice vdip = new VDIP1Device(props, port, lh, this);
				addDevice(vdip);
				cpn = new FT245RDevice(props, port, vdip, lh, this);
				continue;
			}
			if (s.matches("W?H17") || s.matches("W?H8-17")) {
				addDiskDevice(new Z17(props, slot, lh, this, gpp));
				continue;
			}
			if (s.matches("W?H37") || s.matches("W?H8-37")) {
				addDiskDevice(new Z37(props, slot, lh, this));
				continue;
			}
			if (s.matches("W?H47") || s.matches("W?H8-47")) {
			}
			if (s.matches("W?H67") || s.matches("W?H8-67")) {
				addDiskDevice(new MMS77320(props, slot, lh, this));
				continue;
			}
			System.err.format("Unrecognized device \"%s\" in %s\n", s, slot);
		}
		if (mem == null) {
			mem = new H8Memory(props, h8min, h8max, gpp, this);
		}
		if (!has_serial) {
			// H8-4 4-port serial card...
			// console can't be disabled? the rest can.
			if (con == null) {
				INS8250 sp = new INS8250(props, true, "con", 0xe8, 3, this);
				con = sp;
				addDevice(sp);
				con.setModem(VirtualUART.SET_CTS | VirtualUART.SET_DSR);
			}
			if (props.getProperty("lp_dis") == null) {
				addDevice(new INS8250(props, true, "lp", 0xe0, 3, this));
			}
			if (props.getProperty("modem_dis") == null) {
				addDevice(new INS8250(props, true, "modem", 0xd8, 3, this));
			}
			if (props.getProperty("aux_dis") == null) {
				addDevice(new INS8250(props, true, "aux", 0xd0, 3, this));
			}
			has_serial = true;
		}
		if (cpn == null && CPNetDevice.isConfigured(props)) {
			cpn = new CPNetDevice(props, lh, this);
		}
		if (cpn != null) {
			addDevice(cpn);
		}
		return has_serial;
	}

	private boolean configH89(Properties props, LEDHandler lh) {
		String s;
		addDevice(new LegacyH8Ports(props, 0xf0, 2, this));
		addDevice(new LegacyH8Ports(props, 0xfa, 2, this));
		gpp = new H89Gpp(props, this);
		addDevice(gpp);
		String[] memslots = new String[] {"slot_p501", "slot_p502", "slot_p503"};
		String[] ioslots = new String[] {"slot_p504", "slot_p505", "slot_p506",
			"slot_p510", "slot_p511", "slot_p512" // fictitious
			};
		INS8250 sp = new INS8250(props, false, "con", 0xe8, 3, this);
		con = sp;
		addDevice(sp);
		con.setModem(VirtualUART.SET_CTS | VirtualUART.SET_DSR);
		// 'con' is attached via Input/OutputStream i/f.
		boolean mezz = false;
		mem = null;
		for (String slot : memslots) {
			s = props.getProperty(slot);
			if (s == null) {
				continue;
			}
			if (s.equals("MMS77311") || s.equals("MMS77312") ||
						s.equals("WH-88-16")) {
				if (mem != null) {
					System.err.format("WARNING: Conflicting memory " +
							"in %s\n", slot);
				}
				mem = new MMS77311Memory(props, gpp, this);
				continue;
			}
			if (s.equals("MMS77318")) {
				if (mem != null) {
					System.err.format("WARNING: Conflicting memory " +
							"in %s\n", slot);
				}
				mem = new MMS77318Memory(props, gpp, this);
				continue;
			}
			if (s.startsWith("TEST-X2H8")) {
				if (mem != null) {
					System.err.format("WARNING: Conflicting memory " +
							"in %s\n", slot);
				}
				X2H8Memory xm = new X2H8Memory(props, 0x00, gpp, this);
				addDevice(xm);
				mem = xm;
				continue;
			}
			if (s.equals("H89-MEZZANINE")) {
				if (mem != null) {
					System.err.format("WARNING: Conflicting memory " +
							"in %s\n", slot);
				}
				mezz = true;
				H8_512KMemory xm = new H8_512KMemory(props, 0x00, gpp, this);
				addDevice(xm);
				mem = xm;
				continue;
			}
			System.err.format("Unrecognized device \"%s\" in %s\n", s, slot);
		}
		if (mem == null) {
			mem = new H88Memory(props, gpp, this);
		}
		IODevice cpn = null;
		boolean has_serial = false;
		if (mezz) {
			// RAM512K already added
			addDevice(new RTC72421(props, 0xa0));
			// Can't set 'cpn' as WIZ550io is not an IODevice
			addDevice(new H8xSPI(props, "H8xSPI", 0x40, 5, lh, this));
			addDevice(new GIDE_CF(props, "CF", 0x80, lh, this));
			// TODO: any more hardware on mezzanine?
		}
		for (String slot : ioslots) {
			s = props.getProperty(slot);
			if (s == null) {
				continue;
			}
			if (s.equals("H89GIDE")) {
				addDevice(new GIDE_CF(props, null, 0x80, lh, this));
				continue;
			}
			if (s.equals("RTC72421") || s.equals("H89RTCPAR")) {
				// TODO: also a parallel port... and "speed"?
				// TODO: also an ethernet port?
				addDevice(new RTC72421(props, 0xa0));
				continue;
			}
			if (s.equals("MMS77316")) {
				addDiskDevice(new MMS77316(props, slot, lh, this));
				continue;
			}
			if (s.startsWith("MMS77320")) {
				addDiskDevice(new MMS77320(props, slot, lh, this));
				String[] args = s.split("\\s");
				boolean no_serial = false;
				for (String arg : args) {
					if (arg.equals("ns")) no_serial = true;
				}
				if (no_serial) {
					continue;
				}
				if (has_serial) {
					System.err.format("WARNING: %s.%s: ignoring duplicate serial ports\n", slot, s);
					continue;
				}
				addDevice(new INS8250(props, false, "lp", 0xe0, 3, this));
				addDevice(new INS8250(props, false, "modem", 0xd8, 3, this));
				addDevice(new INS8250(props, false, "aux", 0xd0, 3, this));
				has_serial = true;
				continue;
			}
			if (s.equals("H-88-3") || s.equals("HA-88-3") || s.equals("Z-89-11")) {
				if (has_serial) {
					System.err.format("WARNING: %s.%s: ignoring duplicate serial ports\n", slot, s);
					continue;
				}
				addDevice(new INS8250(props, false, "lp", 0xe0, 3, this));
				addDevice(new INS8250(props, false, "modem", 0xd8, 3, this));
				addDevice(new INS8250(props, false, "aux", 0xd0, 3, this));
				has_serial = true;
				continue;
			}
			if (s.equals("H89USBSERIAL")) {
				if (isH8) {
					System.err.format("WARNING: %s: not allowed on H8\n", s);
					continue;
				}
				if (has_serial) {
					System.err.format("WARNING: %s.%s: ignoring duplicate serial ports\n", slot, s);
					continue;
				}
				// re-maps Modem as USB device.
				IODevice vdip = new VDIP1Device(props, 0xd8, lh, this);
				addDevice(vdip);
				cpn = new FT245RDevice(props, 0xd8, vdip, lh, this);
				addDevice(new INS8250(props, false, "lp", 0xe0, 3, this));
				addDevice(new INS8250(props, false, "aux", 0xd0, 3, this));
				has_serial = true;
				continue;
			}
			if (s.equals("MMS77422")) {
				if (isH8) {
					System.err.format("WARNING: %s: not allowed on H8\n", s);
					continue;
				}
				cpn = new MMS77422(props, lh, this);
				continue;
			}
			if (s.matches("[HZ]17") || s.matches("H88-1")) {
				addDiskDevice(new Z17(props, slot, lh, this, gpp));
				continue;
			}
			if (s.matches("[HZ]37") || s.matches("[HZ]89-37")) {
				addDiskDevice(new Z37(props, slot, lh, this));
				continue;
			}
			if (s.matches("[HZ]47")) {
			}
			if (s.matches("[HZ]67")) {
				addDiskDevice(new MMS77320(props, slot, lh, this));
				continue;
			}
			System.err.format("Unrecognized device \"%s\" in %s\n", s, slot);
		}
		if (cpn == null && CPNetDevice.isConfigured(props)) {
			cpn = new CPNetDevice(props, lh, this);
		}
		if (cpn != null) {
			addDevice(cpn);
		}
		return has_serial;
	}

	public void reset() {
		boolean wasRunning = running;
		tracing = false;
		traceCycles = 0;
		traceLow = 0;
		traceHigh = 0;
		// TODO: reset other interrupt state? devices should do that...
		intMask = 0;
		clock = 0;
		stop();
		if (false && wasRunning) {
			System.err.format("backlogNs=%d\n", backlogNs);
		}
		backlogNs = 0;
		cpu.reset();
		mem.reset();
		for (int x = 0; x < devs.size(); ++x) {
			devs.get(x).reset();
		}
		if (wasRunning) {
			start();
		}
	}

	public boolean addDevice(IODevice dev) {
		if (dev == null) {
			System.err.format("NULL I/O device\n");
			return false;
		}
		int base = dev.getBaseAddress();
		int num = dev.getNumPorts();
		if (num < 0) {
			//System.err.format("No ports\n");
			//return false;
			return true;
		}
		for (int x = 0; x < num; ++x) {
			if (ios.get(base + x) != null) {
				System.err.format("Conflicting I/O %02x (%02x)\n", base, num);
				return false;
			}
		}
		devs.add(dev);
		for (int x = 0; x < num; ++x) {
			ios.put(base + x, dev);
		}
		return true;
	}

	public IODevice getDevice(int basePort) {
		IODevice dev = ios.get(basePort);
		return dev;
	}

	public Vector<DiskController> getDiskDevices() {
		return dsks;
	}

	public boolean addDiskDevice(DiskController dev) {
		if (!addDevice(dev)) {
			return false;
		}
		dsks.add(dev);
		return true;
	}

	// These must NOT be called from the thread...
	public void start() {
		stopped = false;
		if (running) {
			return;
		}
		running = true;
		Thread t = new Thread(this);
		t.setPriority(Thread.MAX_PRIORITY);
		t.start();
	}
	public void stop() {
		stopWait.drainPermits();
		if (!running) {
			return;
		}
		running = false;
		// This is safer than spinning, but still stalls the thread...
		try {
			stopWait.acquire();
		} catch (Exception ee) {}
	}
	private void addTicks(int ticks) {
		clock += ticks;
		for (ClockListener lstn : clks) {
			lstn.addTicks(ticks, clock);
		}
		int t = ticks * nanoSecCycle;
		for (TimeListener lstn : times) {
			lstn.addTime(t);
		}
	}

	// I.e. output from the H89 Console Port...
	public InputStream getInputStream() {
		// TODO: single instance
		return (InputStream)new ConsoleInput();
	}

	// I.e. input to the H89 Console Port...
	public OutputStream getOutputStream() {
		// TODO: single instance
		return (OutputStream)new ConsoleOutput();
	}

	// I.e. admin commands to virtual H89...
	public H89Commander getCommander() {
		return (H89Commander)this;
	}

	// TODO: these may be separate classes...

	/////////////////////////////////////////////
	/// Interruptor interface implementation ///
	public int registerINT(int irq) {
		int val = intRegistry[irq & 7]++;
		// TODO: check for overflow (32 bits max?)
		return val;
	}
	public synchronized void raiseINT(int irq, int src) {
		irq &= 7;
		intLines[irq] |= (1 << src);
		intState |= (1 << irq);
		if ((intState & ~intMask) != 0) {
			cpu.setINTLine(true);
		}
	}
	public synchronized void lowerINT(int irq, int src) {
		irq &= 7;
		intLines[irq] &= ~(1 << src);
		if (intLines[irq] == 0) {
			intState &= ~(1 << irq);
			if ((intState & ~intMask) == 0) {
				cpu.setINTLine(false);
			}
		}
	}
	public void blockInts(int msk) {
		intMask |= msk;
		if ((intState & ~intMask) == 0) {
			cpu.setINTLine(false);
		}
	}
	public void unblockInts(int msk) {
		intMask &= ~msk;
		if ((intState & ~intMask) != 0) {
			cpu.setINTLine(true);
		}
	}
	public void triggerNMI() {
		cpu.triggerNMI();
	}
	// TODO: no longer used...
	public void timeout2ms() {
	}
	public void addClockListener(ClockListener lstn) {
		clks.add(lstn);
	}
	public void addTimeListener(TimeListener lstn) {
		times.add(lstn);
	}
	public void addIntrController(InterruptController ctrl) {
		// There really should be only zero or one.
		intrs.add(ctrl);
	}
	public void waitCPU() {
		// Keep issuing clock cycles while stalling execution.
		addTicks(1);
	}
	public boolean isTracing() {
		return tracing;
	}
	public void startTracing(int cy) {
		if (cy > 0) {
			traceCycles = cy;
		} else {
			tracing = true;
		}
	}
	public void stopTracing() {
		traceLow = traceHigh = 0;
		traceCycles = 0;
		tracing = false;
	}

	/////////////////////////////////////////////
	/// OutputStream interface implementation ///
	class ConsoleOutput extends OutputStream {
		public void write(int b) {
			con.put(b, false);
		}
	}

	////////////////////////////////////////////
	/// InputStream interface implementation ///
	class ConsoleInput extends InputStream {
		public int read() {
			return con.take(); // must sleep
		}
		public int available() {
			return con.available();
		}
	}

	/////////////////////////////////////////////
	/// H89Commander interface implementation ///
	public Vector<String> sendCommand(String cmd) {
		// TODO: stop Z80 during command? Or only pause it?
		String[] args = cmd.split("\\s");
		Vector<String> ret = new Vector<String>();
		ret.add("ok");
		Vector<String> err = new Vector<String>();
		err.add("error");
		if (args.length < 1) {
			return ret;
		}
		if (args[0].equalsIgnoreCase("quit")) {
			// Release Z80, if held...
			stop();
			System.exit(0);
		}
		if (args[0].equalsIgnoreCase("trace") && args.length > 1) {
			if (!traceCommand(args, err, ret)) {
				return err;
			}
			return ret;
		}
		try {
			cpuLock.lock(); // This might sleep waiting for CPU to finish 2mS
			if (args[0].equalsIgnoreCase("reset")) {
				reset();
				return ret;
			}
			if (args[0].equalsIgnoreCase("mount")) {
				if (args.length < 3) {
					err.add("syntax");
					err.add(cmd);
					return err;
				}
				// must special-case Cassette...
				GenericRemovableDrive drv = findDrive(args[1]);
				if (drv == null) {
					err.add("nodrive");
					err.add(args[1]);
					return err;
				}
				drv.insertMedia(Arrays.copyOfRange(args, 2, args.length));
				return ret;
			}
			if (args[0].equalsIgnoreCase("unmount")) {
				if (args.length < 2) {
					err.add("syntax");
					err.add(cmd);
					return err;
				}
				GenericRemovableDrive drv = findDrive(args[1]);
				if (drv == null) {
					err.add("nodrive");
					err.add(args[1]);
					return err;
				}
				drv.insertMedia(null);
				return ret;
			}
			if (args[0].equalsIgnoreCase("getdevs")) {
				for (IODevice dev : devs) {
					String nm = dev.getDeviceName();
					if (nm != null) {
						ret.add(nm);
					}
				}
				return ret;
			}
			if (args[0].equalsIgnoreCase("getdisks")) {
				for (DiskController dev : dsks) {
					Vector<GenericRemovableDrive> drvs = dev.getDiskDrives();
					for (GenericRemovableDrive drv : drvs) {
						if (drv != null) {
							ret.add(drv.getDriveName() + "=" + drv.getMediaName());
						}
					}
				}
				return ret;
			}
			if (args[0].equalsIgnoreCase("dump") && args.length > 1) {
				if (args[1].equalsIgnoreCase("core") && args.length > 2) {
					mem.dumpCore(args[2]);
				}
				if (args[1].equalsIgnoreCase("cpu")) {
					ret.add(cpu.dumpDebug());
					ret.add(disas.disas(cpu.getRegPC()) + "\n");
				}
				if (args[1].equalsIgnoreCase("page") && args.length > 2) {
					String s = dumpPage(args);
					if (s == null) {
						err.add("syntax");
						err.addAll(Arrays.asList(args));
						return err;
					}
					ret.add(s);
				}
				if (args[1].equalsIgnoreCase("disas") && args.length > 3) {
					String s = dumpDisas(args);
					if (s == null) {
						err.add("syntax");
						err.addAll(Arrays.asList(args));
						return err;
					}
					ret.add(s);
				}
				if (args[1].equalsIgnoreCase("mach")) {
					ret.add(dumpDebug());
				}
				if (args[1].equalsIgnoreCase("disk") && args.length > 2) {
					IODevice dev = findDevice(args[2]);
					if (dev == null) {
						err.add("nodevice");
						err.add(args[2]);
						return err;
					}
					ret.add(dev.dumpDebug());
				}
				return ret;
			}
			err.add("badcmd");
			err.add(cmd);
			return err;
		} finally {
			cpuLock.unlock();
		}
	}

	private boolean traceCommand(String[] args, Vector<String> err,
			Vector<String> ret) {
		// TODO: do some level of mutexing?
		if (args[1].equalsIgnoreCase("on")) {
			startTracing(0);
		} else if (args[1].equalsIgnoreCase("off")) {
			stopTracing();
		} else if (args[1].equalsIgnoreCase("cycles") && args.length > 2) {
			try {
				traceCycles = Integer.valueOf(args[2]);
			} catch (Exception ee) {}
		} else if (args[1].equalsIgnoreCase("pc") && args.length > 2) {
			// TODO: this could be a nasty race condition...
			try {
				traceLow = Integer.valueOf(args[2], 16);
			} catch (Exception ee) {}
			if (args.length > 3) {
				try {
					traceHigh = Integer.valueOf(args[3], 16);
				} catch (Exception ee) {}
			} else {
				traceHigh = 0x10000;
			}
			if (traceLow >= traceHigh) {
				traceLow = traceHigh = 0;
			}
		} else {
			err.add("unsupported:");
			err.add(args[1]);
			return false;
		}
		return true;
	}

	private GenericRemovableDrive findDrive(String name) {
		for (DiskController dev : dsks) {
			GenericRemovableDrive drv = dev.findDrive(name);
			if (drv != null) {
				return drv;
			}
		}
		return null;
	}

	private IODevice findDevice(String name) {
		for (IODevice dev : devs) {
			if (name.equals(dev.getDeviceName())) {
				return dev;
			}
		}
		return null;
	}

	public void setSpeed(int spd) {
		// spd is in Hz...
		cpuSpeed = spd;
		// TODO: how does this affect the execute loop?
		// Do we need to adjust something? reset the loop?
		cpuCycle2ms = spd / 500;
		nanoSecCycle = 1000000000 / spd;
	}

	////////////////////////////////////////////
	/// GppListener interface implementation ///
	public int interestedBits() { return ssbits; }
	public void gppNewValue(int gpio) {
		ssen = gpp.isSStepEna();
		if (!ssen) {
			lowerINT(2, 0);
			sscnt = 0;
		}
	}

	/////////////////////////////////////////
	/// Computer interface implementation ///

	public int peek8(int address) {
		int val = mem.read(address);
		return val;
	}
	public void poke8(int address, int value) {
		mem.write(address, value);
	}

	// fetch Interrupt Response byte, IM0 (instruction bytes) or IM2 (vector).
	// Implementation must keep track of multi-byte instruction sequence,
	// and other possible state. For IM0, Z80 will call as long as 'intrFetch' is true.
	public int intrResp(Z80State.IntMode mode) {
		if (mode != Z80State.IntMode.IM0) {
			// H89 cannot operate in IM2.(?)  IM1 should never call this.
			return 0; // TODO: What to return in this case?
		}
		int opCode = -1;
		for (InterruptController ctrl : intrs) {
			opCode = ctrl.readDataBus();
			if (opCode >= 0) {
				return opCode;
			}
		}
		// If no other hardware claims intr ack, use H89 hardware.
		// This will always be a single-byte RST instruction.
		int irq = Integer.numberOfTrailingZeros(intState & ~intMask);
		if (irq > 7) {
			// what to do? this is NOT the right action...
			irq = 7;
		}
		// Construct RST instruction form irq.
		opCode = 0xc7 | (irq << 3);
		// TODO: prevent accidental subsequent calls?
		return opCode;
	}

	public void retIntr(int opCode) {
	}

	public int inPort(int port) {
		int val = 0;
		port &= 0xff;
		IODevice dev = ios.get(port);
		if (dev == null) {
			//System.err.format("Undefined Input on port %02x\n", port);
		} else {
			val = dev.in(port);
		}
		return val;
	}
	public void outPort(int port, int value) {
		port &= 0xff;
		IODevice dev = ios.get(port);
		if (dev == null) {
			//System.err.format("Undefined Output on port %02x value %02x\n", port, value);
		} else {
			dev.out(port, value);
		}
	}

	// No longer used...
	public void contendedStates(int address, int tstates) {
		addTicks(tstates);
	}
	// not used?
	public long getTStates() {
		return clock;
	}

	public void breakpoint() {
	}
	public void execDone() {
	}

	public boolean intEnabled() {
		return cpu.isIE();
	}

	public boolean isRunning() {
		return !stopped;
	}

	public void changeSpeed(int mlt, int div) {
		// Z180 only.
		// TODO: anything to do?
		// int spd = (sysSpeed * mlt) / div;
		// setSpeed(spd);
		// ...notifications?
	}

	//////// Runnable /////////
	public void run() {
		String traceStr = "";
		int clk = 0;
		int limit = 0;
		while (running) {
			cpuLock.lock(); // This might sleep waiting for GUI command...
			limit += cpuCycle2ms;
			long t0 = System.nanoTime();
			int traced = 0; // assuming any tracing cancels 2mS accounting
			while (running && limit > 0) {
				int PC = cpu.getRegPC();
				if (ssen) {
					if (sscnt > 0) {
						if (++sscnt == 4) {
							raiseINT(2, 0);
						}
					} else if (mem.read(PC) == 0xfb) {
						// count: EI, RET, <instr>...
						sscnt = 1;
					}
				}
				boolean trace = tracing;
				if (!trace && (traceCycles > 0 ||
						(PC >= traceLow && PC < traceHigh))) {
					trace = true;
				}
				if (trace) {
					++traced;
					traceStr = String.format("{%05d} %04x: %02x %02x %02x %02x " +
						": %02x %04x %04x %04x [%04x] <%02x/%02x>%s",
						clock & 0xffff,
						PC, mem.read(PC), mem.read(PC + 1),
						mem.read(PC + 2), mem.read(PC + 3),
						cpu.getRegA(),
						cpu.getRegBC(),
						cpu.getRegDE(),
						cpu.getRegHL(),
						cpu.getRegSP(),
						intState, intMask,
						cpu.isINTLine() ? " INT" : "");
				}
				clk = cpu.execute();
				if (clk < 0) {
					clk = -clk;
					if (trace) {
						System.err.format("%s {%d} *%s*\n",
							traceStr, clk, cpu.specialCycle());
					}
				} else if (trace) {
					// TODO: collect data after instruction?
					System.err.format("%s {%d} %s\n", traceStr, clk,
						disas.disas(PC));
				}
				limit -= clk;
				if (traceCycles > 0) {
					traceCycles -= clk;
				}
				addTicks(clk);
			}
			cpuLock.unlock();
			if (!running) {
				break;
			}
			long t1 = System.nanoTime();
			if (traced == 0) {
				backlogNs += (2000000 - (t1 - t0));
				t0 = t1;
				if (backlogNs > 10000000) {
					try {
						Thread.sleep(10);
					} catch (Exception ee) {}
					t1 = System.nanoTime();
					backlogNs -= (t1 - t0);
				}
			}
			t0 = t1;
			gpp.trigger2ms();
		}
		stopped = true;
		stopWait.release();
	}

	public String dumpPage(String[] args) {
		String str = "";
		int pg = 0;
		int bnk = -1;
		int i = 2;
		boolean rom = false;
		if (args[i].equalsIgnoreCase("rom")) {
			rom = true;
			++i;
		}
		if (args.length - i > 1) {
			try {
				bnk = Integer.valueOf(args[i++]);
			} catch (Exception ee) {
				return ee.getMessage();
			}
		}
		if (args.length - i < 1) {
			return null;
		}
		try {
			pg = Integer.valueOf(args[i++], 16);
		} catch (Exception ee) {
			return ee.getMessage();
		}
		int adr = pg << 8;
		int end = adr + 0x0100;
		while (adr < end) {
			if (bnk < 0) {
				str += String.format("%04x:", adr);
			} else {
				str += String.format("%d %04x:", bnk, adr);
			}
			for (int x = 0; x < 16; ++x) {
				int c;
				if (bnk < 0) {
					c = mem.read(adr + x);
				} else {
					c = mem.read(rom, bnk, adr + x);
				}
				str += String.format(" %02x", c);
			}
			str += "  ";
			for (int x = 0; x < 16; ++x) {
				int c;
				if (bnk < 0) {
					c = mem.read(adr + x);
				} else {
					c = mem.read(rom, bnk, adr + x);
				}
				if (c < ' ' || c > '~') {
					c = '.';
				}
				str += String.format("%c", (char)c);
			}
			str += '\n';
			adr += 16;
		}
		return str;
	}

	public String dumpDisas(String[] args) {
		int lo = 0;
		int hi = 0;
		String ret = "";
		boolean rom = false;
		int bnk = -1;
		int i = 2;
		if (args[i].equalsIgnoreCase("rom")) {
			rom = true;
			++i;
		}
		if (args.length - i > 2) {
			try {
				bnk = Integer.valueOf(args[i++]);
			} catch (Exception ee) {
				return ee.getMessage();
			}
		}
		try {
			lo = Integer.valueOf(args[i++], 16);
			hi = Integer.valueOf(args[i++], 16);
		} catch (Exception ee) {
			return ee.getMessage();
		}
		for (int a = lo; a < hi;) {
			String d;
			if (bnk < 0) {
				ret += String.format("%04X:", a);
				d = disas.disas(a);
			} else {
				ret += String.format("%d %04X:", bnk, a);
				d = disas.disas(rom, bnk, a);
			}
			int n = disas.instrLen();
			for (int x = 0; x < n; ++x) {
				int b;
				if (bnk < 0) {
					b = mem.read(a + x);
				} else {
					b = mem.read(rom, bnk, a + x);
				}
				ret += String.format(" %02X", b);
			}
			a += n;
			while (n < 4) {
				ret += "   ";
				++n;
			}
			ret += ' ';
			ret += d;
			ret += '\n';
		}
		return ret;
	}

	public String dumpDebug() {
		String ret = gpp.dumpDebug();
		ret += String.format("System Clock %d Hz\n", cpuSpeed);
		ret += String.format("CLK %d", getTStates());
		if (running) {
			ret += " RUN";
		}
		if (stopped) {
			ret += " STOP";
		}
		if (!running && !stopped) {
			ret += " limbo";
		}
		ret += "\n";
		ret += String.format("2mS Backlog = %d nS\n", backlogNs);
		ret += "INT = {";
		for (int x = 0; x < 8; ++x) {
			ret += String.format(" %x", intLines[x]);
		}
		ret += String.format(" } %02x %02x\n", intState, intMask);
		return ret;
	}
}
