// Copyright (c) 2018 Douglas Miller <durgadas311@gmail.com>

import java.util.Vector;
import java.util.Properties;

// H89:
//      |  7  |  6  |  5  |  4  |  3  |  2  |  1  |  0  |
//      |-----|-----|-----|-----|-----|-----|-----|-----|
// F2H  | io1 | io0 |org0 |mem0 |mem1 |     |int1 |int2 |
//      |-----|-----|-----|-----|-----|-----|-----|-----|
// "io0" A.K.A. "ss" (side select)

public class H89Gpp implements GeneralPurposePort {
	private int gpi;
	private int gpo;
	private Interruptor intr;
	private int src;
	private Vector<GppListener> notif;
	static final int sStepBit = (1 << 0);

	public H89Gpp(Properties props, Interruptor intr) {
		this.intr = intr;
		src = intr.registerINT(1);
		gpi = 0;
		notif = new Vector<GppListener>();
		String sw = props.getProperty("sw501");
		if (sw != null) {
			gpi = Integer.valueOf(sw, 2) & 0xff;
		}
		reset();
	}

	public int getOrg0Bit() { return (1 << 5); }
	public int get2mSClockBit() { return (1 << 1); }
	public int getSStepBit() { return sStepBit; }
	public boolean isSStepEna() { return (gpo & sStepBit) != 0; }
	public int getSideSelBit() { return (1 << 6); }
	public int getMEM1Bit() { return (1 << 3); }
	public int getIO1Bit() { return (1 << 7); }

	/// IODevice interface ///

	public void reset() {
		gpo = 0;
		notify(0xff); // might contradict device resets?
	}

	public int getBaseAddress() {
		return 0xf2;
	}
	public int getNumPorts() {
		return 1;
	}
	public String getDeviceName() { return null; }

	public int in(int port) {
		// check port?
		return gpi;
	}
	public void out(int port, int value) {
		// check port?
		value &= 0xff;
		intr.lowerINT(1, src);
		int diff = gpo ^ value;
		gpo = value;
		notify(diff);
	}

	private void notify(int diff) {
		if (diff == 0) {
			return;
		}
		for (GppListener lstn : notif) {
			if ((diff & lstn.interestedBits()) != 0) {
				lstn.gppNewValue(gpo);
			}
		}
	}

	public void addGppListener(GppListener lstn) {
		notif.add(lstn);
		lstn.gppNewValue(gpo);
	}

	public String dumpDebug() {
		String ret = new String();
		ret += String.format("SW501 = %02x, GPIO = %02x\n", gpi, gpo);
		return ret;
	}

	public void trigger2ms() {
		if ((gpo & get2mSClockBit()) != 0) {
			intr.raiseINT(1, src);
		}
	}
}
