// Copyright (c) 2018 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Properties;
import java.util.Random;
import java.io.*;
import java.net.*;
import java.awt.event.*;
import javax.swing.Timer;

public class H8xSPI implements IODevice {
	private String name;
	private int base;
	private int ctrlReg;
	private int stsReg;

	private static final int CTL_SCS0 = 0x01;
	private static final int STS_RDY0 = 0x01;
	private static final int CTL_SCS1 = 0x02;
	private static final int STS_RDY1 = 0x02;
	private static final int CTL_SCS2 = 0x04;
	private static final int STS_RDY2 = 0x04;
	private static final int CTL_SCS3 = 0x08;
	private static final int STS_RDY3 = 0x08;

	private SPIDevice[] devs;
	private int[] stsbits = new int[] {
		STS_RDY0, STS_RDY1, STS_RDY2, STS_RDY3 };
	private int[] ctlbits = new int[] {
		CTL_SCS0, CTL_SCS1, CTL_SCS2, CTL_SCS3 };

	public H8xSPI(Properties props, String id, int port, int irq,
				LEDHandler lh, Interruptor intr) {
		name = id;
		base = port;
		String pfx = id.toLowerCase();
		String s = props.getProperty(pfx + "_port");
		if (s != null) {
			base = Integer.decode(s) & 0xff;
		}
		devs = new SPIDevice[4];
		s = props.getProperty(pfx + "_dev0");
		if (s != null) {
			if (s.matches("[Ww][Ii][Zz].*")) {
				devs[0] = new WIZ550io(props, s, irq, lh, intr);
			} else {
				System.err.format("Unknown device \"%s\" on %s dev0\n",
										s, id);
			}
		}
		s = props.getProperty(pfx + "_dev1");
		if (s != null) {
			// assume NVRAM chip of some sort...
			// typical, expected, value is "25LC512"
			devs[1] = new NVRAM(props, s);
		}
		s = props.getProperty(pfx + "_dev2");
		if (s != null) {
			if (s.matches("[Ss][Dd].*")) {
				devs[2] = new SDCard(props, s, 0, irq, lh, intr);
			} else {
				System.err.format("Unknown device \"%s\" on %s dev2\n",
										s, id);
			}
		}
		s = props.getProperty(pfx + "_dev3");
		if (s != null) {
			if (s.matches("[Ss][Dd].*")) {
				devs[3] = new SDCard(props, s, 1, irq, lh, intr);
			} else {
				System.err.format("Unknown device \"%s\" on %s dev3\n",
										s, id);
			}
		}

		// H8-SPI adapter has an LED for /SCS (any), and PWR.
		// But, adapter card is not visible on front panels.
		// led = lh.registerLED("H8-SPI", LED.Colors.RED);
		reset();
		System.err.format("%s Module at port %02x\n", name, base);
	}

	public void reset() {
		ctrlReg = 0;
		stsReg = 0;
		for (int x = 0; x < 4; ++x) {
			if (devs[x] != null) {
				devs[x].reset();
			}
		}
	}

	public int getBaseAddress() { return base; }
	public int getNumPorts() { return 2; }

	public int in(int port) {
		int val = 0;
		int off = port - base;
		switch (off) {
		case 0:
			val = readData();
			break;
		case 1:
			val = stsReg & 0xff;
			for (int x = 0; x < 4; ++x) {
				if (devs[x] != null && devs[x].status()) {
					val |= stsbits[x];
				}
			}
			break;
		}
		return val;
	}
	public void out(int port, int val) {
		int off = port - base;
		switch (off) {
		case 0:
			writeData(val);
			break;
		case 1:
			setControl(val);
			break;
		}
	}

	private void setControl(int val) {
		int diff = (ctrlReg ^ val);
		ctrlReg = val;
		for (int x = 0; x < 4; ++x) {
			if (devs[x] != null && (diff & ctlbits[x]) != 0) {
				devs[x].scs((ctrlReg & ctlbits[x]) != 0);
			}
		}
	}

	private int readData() {
		int val = 0;
		for (int x = 0; x < 4; ++x) {
			if (devs[x] != null) {
				if ((ctrlReg & ctlbits[x]) != 0) {
					val |= devs[x].miso();
				} else {
					devs[x].sclk(8);
				}
			}
		}
		return val;
	}

	private void writeData(int val) {
		for (int x = 0; x < 4; ++x) {
			if (devs[x] != null) {
				if ((ctrlReg & ctlbits[x]) != 0) {
					devs[x].mosi(val);
				} else {
					devs[x].sclk(8);
				}
			}
		}
	}

	public String getDeviceName() { return name; }

	public String dumpDebug() {
		String ret = String.format("%s Port %02x\n", name, base);
		for (int x = 0; x < 4; ++x) {
			if (devs[x] != null) {
				ret += '\n';
				ret += devs[x].dumpDebug();
			}
		}
		return ret;
	}
}
