// Copyright (c) 2016 Douglas Miller <durgadas311@gmail.com>

import java.io.FileInputStream;
import java.util.Properties;
import java.io.*;

public class MMS77318Memory extends H89Roms implements Memory, GppListener {
	static final int h89_gppBnkSelBit0_c = 0b00100000;
	static final int h89_gppBnkSelBit1_c = 0b00000100;
	static final int h89_gppBnkSelBit2_c = 0b00010000;
	static final int h89_gppBnkSelBits_c = (h89_gppBnkSelBit0_c |
						h89_gppBnkSelBit1_c |
						h89_gppBnkSelBit2_c);
	static final int h89_gppUnlockBits_c = 0b00001100;

	GeneralPurposePort gpp;
	// Organized as [22][8192]...
	private byte[] mem;
	private int lockState;
	private int curBank;
	private int interestedBits;
	private byte[] lockSeq = new byte[] {
		0b00000100,
		0b00001100,
		0b00000100,
		0b00001000,
		0b00001100,
		0b00001000,
	};
	// Organized like [8][8]... Each is base address of 8K chunk in mem[]...
	// "-1" indicate ROM case for read(). "gpp" bits are 4,2,5.
	//
	// bnk gpp descr
	//  0  000 "A" ROM+56K0
	//  1  100 "B" R0+56K0 (48K0+R0,  56K0+R0)
	//  2  001 "C" R2+56K0
	//  3  101 "D" R1+56K0
	//  4  010 "E" 48K2+R0
	//  5  110 "F" 48K1+R0
	//  6  011 "G" 56K2+R0
	//  7  111 "H" 56K1+R0
	private int[] banks = new int[] {
		    -1, 139264, 147456, 155648, 163840, 172032,  49152, 57344, 	// 000 "A"
		131072, 139264, 147456, 155648, 163840, 172032,  49152, 57344, 	// 100 "B"
		 65536, 139264, 147456, 155648, 163840, 172032,  49152, 57344, 	// 001 "C"
		     0, 139264, 147456, 155648, 163840, 172032,  49152, 57344, 	// 101 "D"
		 65536,  73728,  81920,  90112,  98304, 106496,  49152, 57344, 	// 010 "E"
		     0,   8192,  16384,  24576,  32768,  40960,  49152, 57344, 	// 110 "F"
		 65536,  73728,  81920,  90112,  98304, 106496, 114688, 57344, 	// 011 "G"
		     0,   8192,  16384,  24576,  32768,  40960, 122880, 57344, 	// 111 "H"
	};

	public MMS77318Memory(Properties props, GeneralPurposePort gpio, Interruptor intr) {
		super(props, gpio, intr);	// cannot be ROMX
		// TODO: enforce "NOT H8"...
		gpp = gpio;
		interestedBits = h89_gppBnkSelBits_c | h89_gppUnlockBits_c;
		lockState = 1;
		curBank = 0;
		mem = new byte[180224]; // 22x8K - base H89 48K plus 128K add-on.
		// rely entirely on notifications for 'rom' value.
		gpp.addGppListener(this);
	}

	public int read(boolean rom, int bank, int address) {
		int pg = (address >> 13) & 0x0007;
		address &= 0x1fff;
		int x = banks[(bank << 3) + pg]; // actually, base addr of 8K chunk
		if (x < 0) {
			if (address < monMask) {
				return mon.read(address);
			}
			if (address >= 0x1800 && h17 != null) {
				return h17.read(address - 0x1800);
			}
			x = 131072;
		}
		return mem[x + address] & 0xff;
	}
	public int read(int address) {
		return read(false, curBank, address);
	}
	public void write(int address, int value) {
		int pg = (address >> 13) & 0x0007;
		address &= 0x1fff;
		int x = banks[(curBank << 3) + pg]; // actually, base addr of 8K chunk
		if (x < 0) {
			x = 131072;
		}
		mem[x + address] = (byte)value;
	}

	public void reset() {
		interestedBits = h89_gppBnkSelBits_c | h89_gppUnlockBits_c;
		lockState = 1;
		curBank = 0;
	}

	public void dumpCore(String file) {
		try {
			OutputStream core = new FileOutputStream(file);
			core.write(mem);
			core.close();
		} catch (Exception ee) {
			H89Operator.error(null, "Core Dump", ee.getMessage());
		}
	}

	public int interestedBits() {
		return interestedBits;
	}
	public void gppNewValue(int gpio) {
		int bnk = (gpio & h89_gppBnkSelBit0_c) != 0 ? 0x01 : 0; // LSB, GPIO5
		if (lockState == 0) {
			bnk |= (gpio & h89_gppBnkSelBit1_c) != 0 ? 0x02 : 0;  // middle bit
			bnk |= (gpio & h89_gppBnkSelBit2_c) != 0 ? 0x04 : 0;  // MSB
		} else {
			gpio &= h89_gppUnlockBits_c;
			if (gpio == lockSeq[lockState]) {
				++lockState;
				if (lockState >= lockSeq.length) {
					lockState = 0;
					interestedBits = h89_gppBnkSelBits_c;
				}
			} else {
				lockState = 1;
			}
		}
		curBank = bnk; // converted to 1-dimension array index.
	}
}
