// Copyright (c) 2016 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Vector;
import java.util.Properties;
import java.awt.event.*;
import javax.swing.*;

public class MMS77320 extends SASIDriveController
		implements DiskController, ActionListener {
	static final int numDisks_c = 8;
	static final int MMS77320_NumPorts_c = 3;

	static final int DataPort_Offset_c = 0;
	static final int Control0Port_Offset_c = 1;
	static final int StatusPort_Offset_c = 1;
	static final int SwitchPort_Offset_c = 2;
	static final int Control1Port_Offset_c = 2;

	/// Bits set in cmd_Control0Port_c
	static final int ctrl_ResetStart_c = 0x10;
	static final int ctrl_EnableIntReq_c = 0x20;
	static final int ctrl_Select_c = 0x40;
	static final int ctrl_DisableAck_c = 0x80;
	/// Bits set in cmd_Control1Port_c
	static final int ctrl_DriveSel_c = 0x07;

	// Bits in statusReg_m
	static final int sts_Req_c = 0x80;
	static final int sts_POut_c = 0x40;
	static final int sts_Msg_c = 0x20;
	static final int sts_Cmd_c = 0x10;
	static final int sts_Busy_c = 0x08;
	static final int sts_Unused_c = 0x04;
	static final int sts_ActInt_c = 0x02;
	static final int sts_Ack_c = 0x01;

	private String MMS77320_Name = "MMS77320";
	private int BasePort_m = 0xff;		// TBD configured by jumpers
	private int control0Reg_m;
	private int control1Reg_m;
	private int statusReg_m;
	private int switchReg_m;

	GenericSASIDrive[] drives_m;
	GenericSASIDrive curDrive_m;
	private LED[] leds_m;
	private LED curLed_m;
	private javax.swing.Timer timer;
	Interruptor intr;
	private int irq;
	private int src;

	public MMS77320(Properties props, String slot, LEDHandler lh, Interruptor intr) {
		this.intr = intr;
		dataOutReg_m = 0;
		dataInReg_m = 0;
		control0Reg_m = 0;
		control1Reg_m = -1;
		statusReg_m = 0;
		ctrlBus_m = 0;
		curDrive_m = null;
		curLed_m = null;
		timer = new javax.swing.Timer(500, this);

		String s = props.getProperty(slot);
		String[] ss = s.split("\\s");
		MMS77320_Name = ss[0];
		String pfx = MMS77320_Name.toLowerCase();

		BasePort_m = 0x78;
		// TODO: just look at sw501 and make it match?
		// BYTE sw = h89.getIO().in(0xf2);
		// if ((sw & 0x03) == 0x02) port = 0x7c;
		// else if ((sw & 0x0c) == 0x08) port = 0x78;
		// Technically, the user must both strap the card and set SW501 appropriately.
		s = props.getProperty(pfx + "_port");
		if (s != null) {
			if (s.equals("jp1a")) {
				BasePort_m &= ~0x04;
			} else if (s.equals("jp1b")) {
				BasePort_m |= 0x04;
			} else {
				int p = Integer.decode(s);
				// if (p == 0x7c || p == 0x78) {
				BasePort_m = p;
			}
		}
		irq = 0;
		s = props.getProperty(pfx + "_intr");
		if (s != null) {
			if (s.equals("jp2a")) {
				irq = 3;
			} else if (s.equals("jp2b")) {
				irq = 4;
			} else if (s.equals("jp2c")) {
				irq = 5;
			} else {
				int i = Integer.valueOf(s);
				if (i >= 3 && i <= 5) {
					irq = i;
				}
			}
		}
		if (irq == 0) {
			System.err.format("WARNING: no %s_intr specified, using jp2b (INT4)\n", pfx);
			irq = 4;
		}
		src = intr.registerINT(irq);
		switchReg_m = 0;
		s = props.getProperty(pfx + "_dipsw");
		if (s != null) {
			switchReg_m = Integer.valueOf(s, 2) & 0xff;
		}
		String dir = props.getProperty(pfx + "_dir");
		if (dir == null) {
			dir = ".";
		}
		drives_m = new GenericSASIDrive[numDisks_c];
		leds_m = new LED[numDisks_c];
		Arrays.fill(drives_m, null);
		// First identify what drives are installed.
		for (int x = 0; x < numDisks_c; ++x) {
			String prop = String.format(pfx + "_drive%d", x);
			s = props.getProperty(prop);
			if (s != null) {
				ss = s.split("\\s", 2);
				String n = getDriveName(x);
				if (ss.length >= 2) {
					n = ss[1];
				}
				// TODO: handle removable media case...
				GenericSASIDrive drv = GenericSASIDrive.getInstance(props,
						ss[0], n, MMS77320_Name, dir, x);
				if (lh != null) {
					leds_m[x] = lh.registerLED(drv.getDriveName());
					if (drv instanceof Z67IDESASIDrive) {
						((Z67IDESASIDrive)drv).register(lh);
					}
				}
				connectDrive(x, drv);
			}
		}
		System.err.format("%s at port %02x INT%d\n", MMS77320_Name, BasePort_m, irq);
		reset();
	}

	private GenericSASIDrive getCurDrive() {
		// This could be NULL
		return drives_m[control1Reg_m & ctrl_DriveSel_c];
	}

	public Vector<GenericRemovableDrive> getDiskDrives() {
		// No removable drives (?) so nothing returned here...
		Vector<GenericRemovableDrive> drives = new Vector<GenericRemovableDrive>();

		// TODO: handle removeable drives, and return those here...
		// TODO: handle multiple LUNs per controller...
		for (int x = 0; x < numDisks_c; ++x) {
			if (false) {
				drives.add(drives_m[x]);
			}
		}
		return drives;
	}

	public String getDriveName(int index) {
		if (index < 0 || index >= numDisks_c) {
			return null;
		}
		return String.format("%s-%d", MMS77320_Name, index);
	}

	public GenericDiskDrive findDrive(String ident) {
		for (int x = 0; x < numDisks_c; ++x) {
			if (drives_m[x] == null) {
				continue;
			}
			if (ident.equals(drives_m[x].getDriveName())) {
				return drives_m[x];
			}
		}
		return null;
	}

	public int getBaseAddress() { return BasePort_m; }
	public int getNumPorts() { return MMS77320_NumPorts_c; }

	public void reset() {
		if (curDrive_m != null) {
			curDrive_m.deselect(this);
			curLed_m.set(false);
		}
		dataOutReg_m = 0;
		dataInReg_m = 0;
		control0Reg_m = 0;
		control1Reg_m = 0;
		statusReg_m = 0;
		ctrlBus_m = 0;
		curDrive_m = drives_m[0];
		curLed_m = leds_m[0];
		intr.lowerINT(irq, src);
		// TODO: reset all drives?
	}

	public int in(int addr) {
		int offset = addr - BasePort_m;
		int val = 0;
		if (offset == DataPort_Offset_c) {
			val = dataInReg_m;
			if ((ctrlBus_m & GenericSASIDrive.ctl_Out_o_c) != 0) {
				//System.err.format("%s: extraneous data in (%02x)\n",
				//	MMS77320_Name, getStatus(ctrlBus_m));
				return val;
			}
			if (curDrive_m != null) {
				ctrlBus_m |= GenericSASIDrive.ctl_Ack_i_c;
				curDrive_m.ack(this);
			}
			// TODO: side-effects?
		} else if (offset == StatusPort_Offset_c) {
			val = getStatus(ctrlBus_m);
			//System.err.format("in(StatusPort) %02x %s %s %s %s %s\n", val,
			//	(val & sts_Msg_c) != 0 ? "MSG" : "   ",
			//	(val & sts_Req_c) != 0 ? "REQ" : "   ",
			//	(val & sts_Cmd_c) != 0 ? "CMD" : "   ",
			//	(val & sts_POut_c) != 0 ? "OUT" : "   ",
			//	(val & sts_Busy_c) != 0 ? "BSY" : "   "
			//	);
			lowerIntrq();
			// TODO; other side-effects?
		} else if (offset == SwitchPort_Offset_c) {
			val = switchReg_m;
		} else {
			System.err.format("in(Unknown - 0x%02x)\n", addr);
		}
		return val;
	}

	int getStatus(int ctl) {
		int status = 0;
		if ((ctl & GenericSASIDrive.ctl_Req_o_c) != 0) {
			status |= sts_Req_c;
		}
		if ((ctl & GenericSASIDrive.ctl_Out_o_c) != 0) {
			status |= sts_POut_c;
		}
		if ((ctl & GenericSASIDrive.ctl_Msg_o_c) != 0) {
			status |= sts_Msg_c;
		}
		if ((ctl & GenericSASIDrive.ctl_Cmd_o_c) != 0) {
			status |= sts_Cmd_c;
		}
		if ((ctl & GenericSASIDrive.ctl_Busy_o_c) != 0) {
			status |= sts_Busy_c;
		}
		if ((ctl & GenericSASIDrive.ctl_Ack_i_c) != 0) {
			status |= sts_Ack_c;
		}
		// debugss(ssMMS77320, INFO, "raw getStatus %02x (ctl: %02x)\n", status, ctl);
		// TODO: get ACTINT bit...
		// return ~status; // or (status ^ 0xf8)...
		return status;
	}

	public void out(int addr, int val) {
		int offset = addr - BasePort_m;
		if (offset == DataPort_Offset_c) {
			dataOutReg_m = val;
			if ((ctrlBus_m & GenericSASIDrive.ctl_Out_o_c) == 0) {
				//System.err.format("%s: extraneous data out (%02x)\n",
				//	MMS77320_Name, getStatus(ctrlBus_m));
				return;
			}
			if (curDrive_m != null) {
				ctrlBus_m |= GenericSASIDrive.ctl_Ack_i_c;
				curDrive_m.ack(this);
			}
		} else if (offset == Control1Port_Offset_c) {
			val &= 0x0f;
			boolean diff = ((control1Reg_m ^ val) & ctrl_DriveSel_c) != 0;
			control1Reg_m = val;
			if (diff) {
				if (curDrive_m != null) {
					// anything to do to de-select drive?
					curDrive_m.deselect(this);
					curLed_m.set(false);
				}
				curDrive_m = drives_m[control1Reg_m & ctrl_DriveSel_c];
				curLed_m = leds_m[control1Reg_m & ctrl_DriveSel_c];
				// anything else to do to select new drive?
				// actual SELECT is done later, on Control0Port_Offset_c...
			}
		} else if (offset == Control0Port_Offset_c) {
			val &= 0xf0;
			int diffs = (control0Reg_m ^ val);
			control0Reg_m = val;
			if (diffs != 0 && curDrive_m != null) {
				if (control0Reg_m == 0) {
					ctrlBus_m &= ~GenericSASIDrive.ctl_Sel_i_c;
					ctrlBus_m &= ~GenericSASIDrive.ctl_Reset_i_c;
					curDrive_m.run(this);
					timer.removeActionListener(this);
					timer.addActionListener(this);
					timer.restart();
				} else if ((control0Reg_m & ctrl_ResetStart_c) != 0) {
					ctrlBus_m |= GenericSASIDrive.ctl_Reset_i_c;
					curDrive_m.resetSASI(this);
					curLed_m.set(false);
				} else if ((control0Reg_m & ctrl_Select_c) != 0) {
					int temp = dataOutReg_m;
					dataOutReg_m = (1 << (control1Reg_m & ctrl_DriveSel_c));
					ctrlBus_m |= GenericSASIDrive.ctl_Sel_i_c;
					curDrive_m.select(this);
					curLed_m.set(true);
					dataOutReg_m = temp;
				}
			}
		} else {
			System.err.format("out(Unknown - 0x%02x): %02x\n", addr, val);
		}
	}

	public GenericSASIDrive getDrive(int unitNum) {
		if (unitNum < numDisks_c) {
			return drives_m[unitNum];
		}
		return null;
	}

	public boolean connectDrive(int unitNum, GenericSASIDrive drive) {
		boolean retVal = false;
		if (unitNum < numDisks_c) {
			if (drives_m[unitNum] == null) {
				drives_m[unitNum] = drive;
				retVal = true;
			} else {
				System.err.format("MMS77320: drive %d already connect\n", unitNum);
			}
		} else {
			System.err.format("MMS77320: Invalid unit number %d\n", unitNum);
		}
		return retVal;
	}

	boolean removeDrive(int unitNum) {
		return false;
	}

	private void raiseIntrq() {
		if (intrqAllowed()) {
			intr.raiseINT(irq, src);
			return;
		}
	}

	private void lowerIntrq() {
		intr.lowerINT(irq, src);
	}

	public String getDeviceName() { return MMS77320_Name; }

	private boolean intrqAllowed() {
		return (control0Reg_m & ctrl_EnableIntReq_c) != 0;
	}

	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == timer) {
			if (curLed_m != null) {
				curLed_m.set(false);
			}
		}
	}

	public String dumpDebug() {
		String ret = String.format(
			"CTRL-0=%02x     CTRL-1=%02x\n" +
			"DAT-IN=%02x    DAT-OUT=%02x\n" +
			"STATUS=%02x\n",
			control0Reg_m, control1Reg_m,
			dataInReg_m, dataOutReg_m,
			getStatus(ctrlBus_m));
		String sw = Integer.toBinaryString(switchReg_m);
		while (sw.length() < 8) sw = '0' + sw;
		ret += "DIPSW = " + sw + "\n";
		for (int x = 0; x < numDisks_c; ++x) {
			if (drives_m[x] == null) continue;
			ret += drives_m[x].dump();
		}
		return ret;
	}
}
