// Copyright (c) 2016 Douglas Miller <durgadas311@gmail.com>
import java.util.Arrays;
import java.util.Properties;
import java.awt.event.*;
import javax.swing.Timer;

// For reference, the original 77422 board consisted of:
//	Z80A CPU (NEC second-source) @ 4MHz
//	64K DRAM (expandable to 256K)
//	MK3884 (Z80A-SIO equivalent)
//	AM9517 (DMA controller)
//	AM9555? (PIO to Z89 - direct connect to DMA, not accessed by CPU)
//	DRAM Bank select logic
//	SIO-to-Network interface logic (not used in co-processor mode)
// The above would have to be emulated in order to run co-processor mode,
// and a decision made about what to use for '422 monitor ROM image
// (i.e. emulated or actual ROM image).

public class MMS77422 extends ServerDispatch
		implements IODevice, NetworkListener, ActionListener {
	static final int serverId = 0;

	static final int BUFFER_OVERRUN = 512; // any value larger than buffer[]

	static final int dataPortOffset = 0;
	static final int statusPortOffset = 1;

	static final int sts_DataReady = 0x08;
	static final int sts_SendOk = 0x04;
	static final int sts_CmdOverrun = 0x00;
	static final int sts_RespUnderrun = 0x00;
	static final int sts_Error = 0x00;

	static final int mmsLen = NetworkServer.mpayload; // MAGNet header
	static final int ndosLen = NetworkServer.mstart; // includes MAGNet header

	private int clientId;
	private String dir;
	private byte[] buffer = new byte[256 + ndosLen];
	private byte[] D6 = new byte[mmsLen];
	private byte[] netTbl = new byte[mmsLen + 65];
	private byte[] respBuf = null;
	private byte[] pendResp = null;
	private int bufIx;
	private int msgLen;
	private int respLen;
	private boolean intrq = false;
	private int dataReg;
	private int statusReg;
	private int numNodes = 0;

	private int base = 0x78;
	private String name;
	private boolean ledState = false;
	private LED[] led;
	static final int Red = 1;
	static final int Green = 0;
	javax.swing.Timer timer;
	Interruptor intr;
	int irq = 5;
	int src;

	public MMS77422(Properties props, LEDHandler lh, Interruptor intr) {
		super();
		int cid = 0x3f;	// OK default if we have no network connections
		int port = 0x78;
		String s;
		int no = 0;

		s = props.getProperty("mms77422_port");
		if (s != null) {
			try {
				int p = Integer.decode(s);
				if (p == 0x78 || p == 0x7c) {
					port = p;
				} else {
					System.err.format("Invalid MMS77422 port %02x, using default\n", p);
				}
			} catch (Exception ee) {
				System.err.format( "Invalid MMS77422 port number \"%s\", using default\n", s);
			}
		}

		s = props.getProperty("mms77422_intr");
		if (s != null) {
			try {
				int i = Integer.valueOf(s);
				if (i == 3 || i == 4 || i == 5) {
					irq = i;
				} else {
					System.err.format("Invalid MMS77422 intr %d, using default\n", i);
				}
			} catch (Exception ee) {
				System.err.format( "Invalid MMS77422 intr number \"%s\", using default\n", s);
			}
		}
		s = props.getProperty("mms77422_clientid");
		if (s != null) {
			try {
				int c = Integer.decode(s);
				if (c > 0x00 && c < 0x40) {
					cid = c;
				} else {
					System.err.format("Invalid MMS77422 client ID \"%s\"\n", s);
				}
			} catch (Exception ee) {
				System.err.format("Invalid MMS77422 client ID \"%s\"\n", s);
			}
		}
		System.err.format("Creating MMS77422 device at port %02x, intr %d, client ID %02x\n",
			port, irq, cid);
		name = String.format("MAGNet-%02x", cid);
		// TODO: acquire two LEDs, "online" and "activity"? GREEN and AMBER?
		led = lh.registerLEDs(getDeviceName(), 2, new LED.Colors[]{LED.Colors.GREEN,LED.Colors.RED});
		timer = new javax.swing.Timer(500, this);
		src = intr.registerINT(irq);

		base = port;
		clientId = cid;
		this.intr = intr;
		bufIx = 0;
		msgLen = 0;
		respLen = 0;

		Arrays.fill(netTbl, (byte)0);
		putCode(netTbl, 0x30);
		putBC(netTbl, 0x1000 | clientId);
		// TODO: differentiate between CP/M and CP/NET?
		netTbl[NetworkServer.mpayload + 1 + clientId] = (byte)NetworkServer.tcpnet;
		numNodes = 1;

		Arrays.fill(D6, (byte)0);
		putCode(D6, 0x38);

		// TODO: let ServerDispatch pre-fill netTbl?
		install_ServerDispatch(props, "mms77422_", cid, no, 64, this);

		reset();
	}

	public int getBaseAddress() { return base; }
	public int getNumPorts() { return 2; }

	public void reset() {
		// Only the H89-interface is reset...
		// the 77422/network is not affected.
		statusReg = sts_SendOk;
		dataReg = 0;
		lowerIntrq();
		bufIx = 0;
		msgLen = 0;
		respLen = 0;
		pendResp = null;
		respBuf = null;
	}

	private void raiseIntrq() {
		intr.raiseINT(irq, src);
		intrq = true;
	}

	private void lowerIntrq() {
		intr.lowerINT(irq, src);
		intrq = false;
	}

	private void startMsg(byte[] msg, int len) {
		respBuf = msg;
		respLen = len;
		bufIx = 0;
		if (bufIx < respLen) {
			dataReg = respBuf[bufIx++] & 0xff;
			statusReg |= sts_DataReady;
		}
	}

	private void nextByte() {
		if (bufIx < respLen) {
			boolean hdr = bufIx < mmsLen;
			dataReg = respBuf[bufIx++] & 0xff;
			statusReg |= sts_DataReady;
			if (bufIx >= respLen) { 
				//dumpMsg("Done", respBuf, respLen);
				respLen = 0;
				bufIx = 0;
				raiseIntrq();
			} else if (hdr && bufIx >= mmsLen) {
				raiseIntrq();
			}
		} else if (pendResp != null) {
			startMsg(pendResp, mmsLen + getBC(pendResp));
			pendResp = null;
			//dumpMsg("Pend", respBuf, respLen);
		}
	}

	private void dumpMsg(String tag, byte[] buf, int len) {
		String str = String.format("%s: [%d] %02x %04x %04x %04x :",
			tag, len, getCode(buf), getBC(buf), getDE(buf),
			getHL(buf));
		for (int x = mmsLen; x < len; ++x) {
			str += String.format(" %02x", buf[x]);
		}
		System.err.format("%s\n", str);
	}

	private void processMsg() {
		//dumpMsg("Got", buffer, msgLen);
		int cd = getCode(buffer);
		if (cd == 0x00 || cd == 0x20) {
			buffer[NetworkServer.mDE] = (byte)clientId;
			if (cd == 0x00) {
				buffer[NetworkServer.mDEh] =
					buffer[NetworkServer.mdid];
			}
			// C1 - standard CP/NET message, or
			// B1 - BOOT
			startSend();
			int len;
			// let consumer sort out MAGNet format
			byte[] rsp = sendMagNETMsg(buffer, msgLen);
			// setup "send status"
			startMsg(D6, mmsLen);
			if (rsp != null) {
				// must return "send status" first...
				// unless this *is* send status (error)...
				if (getCode(rsp) == 0x38) {
					startMsg(rsp, mmsLen);
				} else {
					pendResp = rsp;
					//dumpMsg("Resp", rsp, mmsLen + getBC(rsp));
				}
			}
		} else if (cd == 0x30) {
			// setup status.
			++netTbl[NetworkServer.mpayload];
			netTbl[NetworkServer.mpayload] &= 0x3f;
			startMsg(netTbl, netTbl.length);
		} else {
			// E1 - execute program on co-processor...
			// A1 - enter monitor mode...
			// just ignore? no response required?
			bufIx = 0;
		}
		// must not alter bufIx here, startMsg() already set it up...
		msgLen = 0;
	}

	private void online() {
		led[Green].set(true);
	}
	private void offline() {
		led[Green].set(false);
		led[Red].set(false);
		ledState = false;
	}

	private void startSend() {
		if (!ledState) {
			led[Red].set(true);
			ledState = true;
		}
		timer.removeActionListener(this);
		timer.addActionListener(this);
		timer.restart();
	}

	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == timer) {
			timer.removeActionListener(this);
			led[Red].set(false);
			ledState = false;
		}
	}

	public int in(int adr) {
		int off = adr - base;
		int val = 0x00;

		if (off == statusPortOffset) {
			if ((statusReg & sts_DataReady) != 0) {
				// the client has work to do.
				return statusReg;
			}
			// must be waiting for a response, check and see...
			// Should never get here for synchronous server handlers
			// (e.g. HostFileBdos).
			byte[] rsp = checkRecvMsg((byte)clientId);
			//
			// TODO: could we reach here while a response
			// is still pending? If we allow unsolicited messages,
			// need to cope with possible conflict.
			//
			if (rsp != null) {
				// This includes any failures or errors,
				// checkRecvMsg(el al.) has formatted correctly.
				// D0 and D6 do not land here.
				pendResp = null;
				startMsg(rsp, getBC(rsp) + mmsLen);
			}
			return statusReg;
		}
		if (off == dataPortOffset) {
			val = dataReg;
			statusReg &= ~sts_DataReady;
			nextByte();
			return val;
		}
		System.err.format("Invalid port address %02x\n", adr);
		return val;
	}

	public void out(int adr, int val) {
		int off = adr - base;

		if (off == statusPortOffset) {
			// IACK - Interrupt Acknowledge
			lowerIntrq();
			return;
		}
		if (off == dataPortOffset) {
			if (respLen > 0) {
				// TODO: is this allowed on '422?
				// response was not yet consumed.
				// force error?
				msgLen = BUFFER_OVERRUN;
				return;
			}
			if (bufIx < mmsLen) {
				buffer[bufIx++] = (byte)val;
				if (bufIx < mmsLen) {
					return;
				}
				// some messages have no data, need to
				// recognize those and not look at BC.
				// Also need to fix-up BC for others.
				int cd = getCode(buffer);
				if (cd == 0x00) {
					msgLen = mmsLen + getBC(buffer);
					return; // more data follows
				} else if (cd == 0x20) {
					putBC(buffer, 0);
					// ready to process... fall-through
				} else {
					// TODO: do any of these have payload?
					// ready to process... fall-through
				}
				processMsg();
				return;
			}
			if (bufIx < msgLen) {
				buffer[bufIx++] = (byte)val;
				if (bufIx >= msgLen) {
					// we have something to do...
					processMsg();
				}
				// don't do anything, just wait for receiver to execute command?
				return;
			}
			// too many bytes - probably some lost/aborted message.
			// need to remember error condition for later.
			System.err.format("Command buffer overrun %d/%d\n", bufIx, msgLen);
			msgLen = BUFFER_OVERRUN;
			return;
		}
		System.err.format("Invalid port address %02x\n", adr);
	}

	public String getDeviceName() { return name; }

	public void addNode(int nodeId, int type) {
		netTbl[NetworkServer.mpayload + 1 + nodeId] = (byte)type;
		++numNodes;
		online();
	}

	public void dropNode(int nodeId) {
		netTbl[NetworkServer.mpayload + 1 + nodeId] = (byte)NetworkServer.toffline;
		--numNodes;
		if (numNodes < 2) {
			offline();
			if (numNodes < 1) {
				System.err.format("Node count went negative\n");
				numNodes = 1;
			}
		}
	}

	public String dumpDebug() {
		String str = String.format("Status: %s%s%s [%d]\n",
			msgLen == BUFFER_OVERRUN ? "CmdOverrun " : "-",
			respLen < 0 ? "RespUnderrun " : "-",
			respLen > 0 ? "DataReady " : "-",
			bufIx);
		str += String.format("[%d] %02x %04x %04x %04x : %02x %02x\n",
			msgLen,
			getCode(buffer), getBC(buffer), getDE(buffer),
			getHL(buffer),
			buffer[NetworkServer.mcode] & 0xff,
			buffer[NetworkServer.mdid] & 0xff);
		if (respLen > 0 && respBuf != null) {
			str += String.format("[%d] %02x %04x %04x %04x : %02x %02x\n",
				respLen,
				getCode(respBuf), getBC(respBuf), getDE(respBuf),
				getHL(respBuf),
				respBuf[NetworkServer.mcode] & 0xff,
				respBuf[NetworkServer.mdid] & 0xff);
		}
		return str;
	}
}
