// Copyright (c) 2021 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Properties;
import java.util.Vector;
import java.util.Map;
import java.util.HashMap;
import java.awt.event.*;
import java.io.*;

// Message format:
//    BI BD [maps...] LN str... 00
// where BI = BIOS page number - start of BIOS in memory
//       BD = BDOS page number - start of BDOS in memory
//       [maps...] are optional CFGTBL entries (two bytes ea):
//            0x80+rd, ld
//            rd = remote device, 0-15=A:-P:, or LST: 0-15 if ld=17
//            ld = local device, 0-15=A:-P:, 17=LST:
//       LN is length of string, including NUL (00).
//
// Search order (XX = client node ID):
// 'tag' specifed:
//	.../sniosXXtag.spr	.../ndosXXtag.spr
//	.../sniostag.spr	.../ndostag.spr
// No 'tag':
//	.../sniosXX.spr		.../ndosXX.spr
//	.../snios.spr		.../ndos.spr
//
// It is not necessary that both snios and ndos match the same pattern
// (it may be advantagous to have different snios files for each client
//  but use the same ndos for all). However, if a 'tag' is specified then
// there must be a file name containing that tag.

public class NetBootLdr implements NetBoot {
	static final int BIOSPG = 0;
	static final int BDOSPG = 1;
	static final int MAPS = 2;
	static final int RECLEN = 128;

	private String dir;
	private String def;
	private int nid;
	private int sid;

	private int bios_pg;
	private int bdos_pg;
	private int snios_pg;
	private int ndos_pg;

	private int entry;
	private int cfgtbl;
	private int curdma;
	private int reccnt;
	private boolean org0;
	private File lastFile;	// File obj of last loadSpr()
	private String ldmsg;
	private String sniosFile;
	private String ndosFile;
	private byte[] image;
	private Map<Integer,Integer> maps;

	public NetBootLdr(String dir, String def, byte[] msg) throws Exception {
		maps = new HashMap<Integer,Integer>();
		this.dir = dir;
		this.def = def;
		nid = msg[NetworkServer.SID] & 0xff;
		sid = msg[NetworkServer.DID] & 0xff; // our ID
		bios_pg = (msg[NetworkServer.DAT+BIOSPG] & 0xff) << 8;
		bdos_pg = (msg[NetworkServer.DAT+BDOSPG] & 0xff) << 8;
		int n = (msg[NetworkServer.SIZ] & 0xff) + 1;
		// assert(n >= MAPS);
		int x = MAPS;
		while (x + 1 < n && (msg[NetworkServer.DAT + x] & 0x80) != 0) {
			int rd = msg[NetworkServer.DAT + x] & 0x0f;
			int ld = msg[NetworkServer.DAT + x + 1] & 0xff;
			maps.put(ld, (rd << 8) | sid);
			x += 2;
		}
		// msg[x] = string length, msg[x+1] = string
		String tag = "";
		if (n - x > 2) {
			tag = new String(msg, NetworkServer.DAT + x + 1, n - x - 2).toLowerCase();
		}
		byte[] snios = loadSpr(tag, "snios");
		sniosFile = lastFile.getAbsolutePath();
		cfgtbl = getCfgTbl(snios);
		setCfgTbl(maps, snios);
		fixCfgTbl(snios);
		byte[] ndos = loadSpr(tag, "ndos");
		ndosFile = lastFile.getAbsolutePath();
		snios_pg = bdos_pg - lengthSpr(snios);
		ndos_pg = snios_pg - lengthSpr(ndos);
		relocSpr(snios_pg, snios);
		relocSpr(ndos_pg, ndos);
		// TODO: combine into one...
		image = new byte[lengthSpr(snios) + lengthSpr(ndos)];
		System.arraycopy(ndos, 0x0100, image, 0, lengthSpr(ndos));
		System.arraycopy(snios, 0x0100, image, lengthSpr(ndos), lengthSpr(snios));

		ldmsg = String.format(	"BIOS        %04X %04X\r\n" +
					"BDOS        %04X %04X\r\n" +
					"SNIOS   SPR %04X %04X\r\n" +
					"NDOS    SPR %04X %04X\r\n" +
					"\n%dK TPA$",
				bios_pg, 0x10000 - bios_pg,
				bdos_pg, bios_pg - bdos_pg,
				snios_pg, bdos_pg - snios_pg,
				ndos_pg, snios_pg - ndos_pg,
				ndos_pg >> 10
			);
		// assert(ldmsg.length() <= 128);
		entry = ndos_pg + 3;
		org0 = false;
		System.err.format("Client %02x loading CP/NET from %s, %s\n",
					nid, ndosFile, sniosFile);
	}

	// Raw length, often needs to be rounded-up to full page.
	private int sizeSpr(byte[] img) {
		int n = (img[1] & 0xff) | ((img[2] & 0xff) << 8);
		return n;
	}
	private int lengthSpr(byte[] img) {
		int n = sizeSpr(img);
		n = (n + 0xff) & 0xff00; // round up to page
		return n;
	}

	// For now, only support single-segment SPR files
	private byte[] loadSpr(String tag, String base) throws Exception {
		byte[] img = null;
		File file = null;

		if (tag.length() > 0) {
			file = new File(String.format("%s/%s%02X%s.spr",
							dir, base, nid, tag));
			if (!file.exists()) {
				file = new File(String.format("%s/%s%s.spr",
							dir, base, tag));
			}
		} else {
			file = new File(String.format("%s/%s%02X.spr", dir, base, nid));
			if (!file.exists()) {
				file = new File(String.format("%s/%s.spr", dir, base));
			}
		}
		lastFile = file;
		// Just let new FileInputStream() throw "file not found"...
		FileInputStream f = new FileInputStream(file);
		img = new byte[f.available()];
		f.read(img);
		f.close();
		return img;
	}

	private void relocSpr(int adr, byte[] img) {
		int ln = sizeSpr(img);
		int bm = ln + 0x0100;
		int pg = (adr >> 8);
		for (int x = 0; x < ln; ++x) {
			int bit = (x & 7);
			int byt = (x >> 3);
			int b = img[bm + byt] & 0xff;
			if (b == 0) {
				x += 7; // 8 total, w/++
				continue;
			}
			if (((b << bit) & 0x80) == 0) {
				continue;
			}
			int hi = img[0x0100 + x] & 0xff;
			// TODO: handle special addresses...
			img[0x0100 + x] = (byte)(hi + pg);
		}
	}

	private int getCfgTbl(byte[] img) {
		int adr = 0x0106; // location of CFGTBL jump vector
		if ((img[adr] & 0xff) != 0xc3) {	// JMP ... ?
			return -1;
		}
		adr = (img[adr + 1] & 0xff) | ((img[adr + 2] & 0xff) << 8);
		adr += 0x0100; // translate address to offset
		if ((img[adr] & 0xff) != 0x21) {	// LXI H,... ?
			return -1;
		}
		return (img[adr + 1] & 0xff) | ((img[adr + 2] & 0xff) << 8);
	}

	private void setCfgTbl(Map<Integer,Integer> maps, byte[] img) {
		if (cfgtbl < 0) return;
		int devtbl = cfgtbl + 2; // drive/device table
		devtbl += 0x0100; // translate address to offset
		for (int x = 0; x < 18; ++x) {
			if (!maps.containsKey(x)) continue;
			int map = maps.get(x);
			int c = devtbl + (x * 2);
			img[c] = (byte)((map >> 8) | 0x80);
			img[c + 1] = (byte)map;
		}
	}

	private void fixCfgTbl(byte[] img) {
		if (cfgtbl < 0) return;
		int devtbl = cfgtbl + 2; // drive/device table
		devtbl += 0x0100; // translate address to offset
		// fixup A:-P:,CON:,LST:
		for (int x = 0; x < 16+2; ++x) {
			int i = x * 2;
			if ((img[devtbl + i] & 0x80) != 0 &&
				(img[devtbl + i + 1] & 0xff) == 0xff) {
				img[devtbl + i + 1] = (byte)(sid);
			}
		}
	}

	public boolean hasRes() { return true; }
	public boolean hasBnk() { return false; }

	public void setLoadMsg(byte[] resp, int start, int max) {
		byte[] lm = ldmsg.getBytes();
		System.arraycopy(lm, 0, resp, start, lm.length);
	}

	public int startResAddr() {
		curdma = 0;
		reccnt = image.length / RECLEN;
		return ndos_pg;
	}

	public int nextResRec(byte[] resp, int start, int max) {
		System.arraycopy(image, curdma, resp, NetworkServer.DAT, RECLEN);
		curdma += RECLEN;
		--reccnt;
		return reccnt;
	}

	public int startBnkAddr() {
		// must never be called!
		return 0;
	}

	public int nextBnkRec(byte[] resp, int start, int max) {
		// must never be called!
		return -1;
	}

	public int entryAddr() {
		return entry;
	}

	public boolean needsOrg0() {
		return org0;
	}
}
