// Copyright (c) 2019 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Properties;
import java.util.Map;
import java.util.HashMap;
import java.util.Vector;
import java.awt.*;
import java.awt.event.*;
import javax.swing.Timer;
import javax.swing.*;
import java.io.*;
import java.nio.file.attribute.BasicFileAttributes;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.FileSystems;
import java.nio.file.StandardCopyOption;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.FileVisitResult;
import java.util.TimeZone;
import java.util.Calendar;

// Commands requiring special handling in IPH mode:
//
// 	OPR 0x20 <filename> [ 0x20 0xNN 0xNN 0xNN 0xNN ] 0x0d
// 	OPW 0x20 <filename> [ 0x20 0xNN 0xNN 0xNN 0xNN ] 0x0d
// 	WRF 0x20 0xNN 0xNN 0xNN 0xNN 0x0d
// 	RDF 0x20 0xNN 0xNN 0xNN 0xNN 0x0d
// 	SEK 0x20 0xNN 0xNN 0xNN 0xNN 0x0d

public class VDIP1Device
		implements IODevice, ActionListener {

	static final int dataPortOffset = 1;
	static final int statusPortOffset = 2;
	static final int sts_DataReady = 0x08;
	static final int sts_TxReady = 0x04;

	static final int CMD_SCS = 0x10;
	static final int CMD_ECS = 0x11;
	static final int CMD_IPA = 0x90;
	static final int CMD_IPH = 0x91;
	static final int CMD_SBD = 0x14;
	static final int CMD_FWV = 0x13;
	static final int CMD_E = 0x45;
	static final int CMD_e = 0x65;
	static final int CMD_DIR = 0x01;
	static final int CMD_CD = 0x02;
	static final int CMD_RD = 0x04;
	static final int CMD_DLD = 0x05;
	static final int CMD_MKD = 0x06;
	static final int CMD_DLF = 0x07;
	static final int CMD_WRF = 0x08;
	static final int CMD_OPW = 0x09;
	static final int CMD_CLF = 0x0a;
	static final int CMD_RDF = 0x0b;
	static final int CMD_REN = 0x0c;
	static final int CMD_OPR = 0x0e;
	static final int CMD_SEK = 0x28;
	static final int CMD_FS = 0x12;
	static final int CMD_FSE = 0x93;
	static final int CMD_IDD = 0x0f;
	static final int CMD_IDDE = 0x94;
	static final int CMD_DSN = 0x2d;
	static final int CMD_DVL = 0x2e;
	static final int CMD_DIRT = 0x2f;
	// VDFC commands
	static final int CMD_CPF = 0x20;
	static final int CMD_CPS = 0x21;
	static final int CMD_IMF = 0x25;
	static final int CMD_IMS = 0x26;
	static final int CMD_A = 0x413a;	// "A:" actually
	static final int CMD_B = 0x423a;	// "B:" actually

	private static Map<String, Integer> cmds;

	static final int ERR_BC = 1;
	static final int ERR_CF = 2;
	static final int ERR_DF = 3;
	static final int ERR_FI = 4;
	static final int ERR_RO = 5;
	static final int ERR_FO = 6;
	static final int ERR_NE = 7;
	static final int ERR_FN = 8;
	static final int ERR_NU = 9;
	static final int ERR_ND = 10;
	private static Vector<byte[]> errs;
	private static Vector<String> errors;

	private int basePort;
	private java.util.concurrent.LinkedBlockingDeque<byte[]> fifi;
	private TimeZone tz = TimeZone.getDefault();

	private byte[] cmdBuf;
	private int cmdIdx;	// loading index (end of command)
	private int cmdPix;	// parsing index
	private int cmdErr;	// for WRF (over data phase)
	private byte[] dataBuf;
	private int dataIdx;
	private int inIdx;
	private byte[] inBuf;
	private int bufoff = 0;
	private int buflen = 0;
	private int last = 0;

	private LED led;
	javax.swing.Timer timer;

	private String fw;
	private boolean vdfc;
	private boolean cmd;
	private boolean bin;
	private boolean hex;
	private File[] dir;
	private File[] cwd;
	private int uniq = 0;
	private int idx;
	private JPanel pnl;
	private JPopupMenu pu;
	private JMenuItem[] pus;
	private RandomAccessFile ofile;
	private long ofTime;	// -1 for "none"
	private File ofFile;
	private boolean ofWrite;

	static {
		cmds = new HashMap<String, Integer>();
		cmds.put("SCS", CMD_SCS);
		cmds.put("ECS", CMD_ECS);
		cmds.put("IPA", CMD_IPA);
		cmds.put("IPH", CMD_IPH);
		cmds.put("SBD", CMD_SBD);
		cmds.put("FWV", CMD_FWV);
		cmds.put("E", CMD_E);	// handle special?
		cmds.put("e", CMD_e);	// handle special?
		cmds.put("DIR", CMD_DIR);
		cmds.put("CD", CMD_CD);
		cmds.put("RD", CMD_RD);
		cmds.put("DLD", CMD_DLD);
		cmds.put("MKD", CMD_MKD);
		cmds.put("DLF", CMD_DLF);
		cmds.put("WRF", CMD_WRF);
		cmds.put("OPW", CMD_OPW);
		cmds.put("CLF", CMD_CLF);
		cmds.put("RDF", CMD_RDF);
		cmds.put("REN", CMD_REN);
		cmds.put("OPR", CMD_OPR);
		cmds.put("SEK", CMD_SEK);
		cmds.put("FS", CMD_FS);
		cmds.put("FSE", CMD_FSE);
		cmds.put("IDD", CMD_IDD);
		cmds.put("IDDE", CMD_IDDE);
		cmds.put("DSN", CMD_DSN);
		cmds.put("DVL", CMD_DVL);
		cmds.put("DIRT", CMD_DIRT);

		errs = new Vector<byte[]>();
		errs.add(0, "OK".getBytes());
		errs.add(ERR_BC, "BC".getBytes());
		errs.add(ERR_CF, "CF".getBytes());
		errs.add(ERR_DF, "DF".getBytes());
		errs.add(ERR_FI, "FI".getBytes());
		errs.add(ERR_RO, "RO".getBytes());
		errs.add(ERR_FO, "FO".getBytes());
		errs.add(ERR_NE, "NE".getBytes());
		errs.add(ERR_FN, "FN".getBytes());
		errs.add(ERR_NU, "NU".getBytes());
		errs.add(ERR_ND, "ND".getBytes());

		errors = new Vector<String>();
		errors.add(0, "Success");
		errors.add(ERR_BC, "Bad Command");
		errors.add(ERR_CF, "Command Failed");
		errors.add(ERR_DF, "Disk Full");
		errors.add(ERR_FI, "Invalid");
		errors.add(ERR_RO, "Read Only");
		errors.add(ERR_FO, "File Open");
		errors.add(ERR_NE, "Dir Not Empty");
		errors.add(ERR_FN, "Filename Invalid");
		errors.add(ERR_NU, "No Upgrade");
		errors.add(ERR_ND, "No Disk");
	}

	public VDIP1Device(Properties props, int base, LEDHandler lh, Interruptor intr) {
		super();
		basePort = base;
		fifi = new java.util.concurrent.LinkedBlockingDeque<byte[]>();
		cmdBuf = new byte[64];
		dataBuf = null;
		cmdIdx = 0;
		dataIdx = 0;
		led = lh.registerLED(getDeviceName());
		Container c = led.getParent();
		if (c instanceof JPanel) {
			pnl = (JPanel)c;
		}
		timer = new javax.swing.Timer(500, this);
		dir = new File[2];
		cwd = new File[2];
		String s = props.getProperty("vdip1device_fw");
		if (s != null) {
			fw = s;
		} else {
			fw = "03.69VDAP";
		}
		vdfc = fw.matches(".*VDFC.*");
		JMenuItem mi = new JMenuItem(getDeviceName(), KeyEvent.VK_P);
		mi.addActionListener(this);
		lh.setMenuItem(getDeviceName(), mi);
		if (vdfc) {
			cmds.put("A:", CMD_A);
			cmds.put("B:", CMD_B);
			cmds.put("IMS", CMD_IMS);
			cmds.put("IMF", CMD_IMF);
			cmds.put("CPS", CMD_CPS);
			cmds.put("CPF", CMD_CPF);
			pus = new JMenuItem[2];
			pu = new JPopupMenu("VDIP1");
			pus[0] = new JMenuItem("P1", KeyEvent.VK_1);
			pus[0].addActionListener(this);
			pu.add(pus[0]);
			pus[1] = new JMenuItem("P2", KeyEvent.VK_2);
			pus[1].addActionListener(this);
			pu.add(pus[1]);
		}
		setDir(props, 1, "vdip1device_dir");
		setDir(props, 1, "vdip1device_dir2");
		setDir(props, 0, "vdip1device_dir1");
		System.err.format("VDIP1 device at port %02x f/w %s\n",
			basePort, fw);
		reset();
	}

	public int getBaseAddress() { return basePort; }
	public int getNumPorts() { return 0; } // addDevice() but no ports (slave to FT345R)
	public String getDeviceName() { return "VDIP1"; }

	private void setDir(Properties props, int ix, String p) {
		String s = props.getProperty(p);
		if (s != null) {
			setDir(ix, new File(s));
		}
	}

	private void updateTip() {
		if (pnl == null) {
			return;
		}
		if (vdfc) {
			String s = "<HTML>";
			String p;
			if (dir[0] != null) {
				p = "P1=" + dir[0].getName();
			} else {
				p = "P1=EMPTY";
			}
			pus[0].setText(p);
			s += p;
			s += "<BR>";
			if (dir[1] != null) {
				p = "P2=" + dir[1].getName();
			} else {
				p = "P2=EMPTY";
			}
			pus[1].setText(p);
			s += p;
			s += "</HTML>";
			pnl.setToolTipText(s);
		} else {
			if (dir[1] != null) {
				pnl.setToolTipText(dir[1].getName());
			} else {
				pnl.setToolTipText("EMPTY");
			}
		}
	}

	private void setDir(int ix, File d) {
		if (ofile != null) {
			try {
				ofile.close();
			} catch (Exception ee) {}
			ofile = null;
		}
		// TODO: trigger events for VDPS f/w:
		// "Device removed from USB Port 1"
		// "Device inserted in USB Port 1"
		if (d == null) {
			dir[ix] = null;
			cwd[ix] = null;
			if (vdfc && idx == ix && dir[idx ^ 1] != null) {
				idx ^= 1;
			}
			updateTip();
			return;
		}
		if (!d.exists()) try {
			boolean did = d.mkdir();
			if (!did) {
				return;
			}
		} catch(Exception ee) {
			return;
		}
		dir[ix] = d;
		cwd[ix] = d;
		if (vdfc && dir[idx] == null) {
			idx = ix;
		}
		updateTip();
	}

	private void sendPrompt() {
		if (bin) {
			sendBack(">\r");
		} else if (vdfc) {
			sendBack(String.format("%c:\\>\r", (char)('A' + idx)));
		} else {
			sendBack("D:\\>\r");
		}
	}

	private boolean ledState = false;

	private void startLED() {
		if (!ledState && led != null) {
			led.set(true);
			ledState = true;
		}
		timer.removeActionListener(this);
		timer.addActionListener(this);
		timer.restart();
	}

	public void reset() {
		//super.reset();
		// clear registers...
		fifi.clear();
		if (ofile != null) {
			try {
				ofile.close();
			} catch (Exception ee) {}
			ofile = null;
		}
		cmd = true;	// receiving command
		bin = false;	// extended command mode
		hex = true;	// binary number mode
		cmdIdx = 0;
		dataIdx = 0;
		if (ofile != null) {
			try {
				ofile.close();
			} catch (Exception ee) {}
			ofile = null;
		}
		if (vdfc && dir[0] != null) {
			idx = 0;
		} else {
			idx = 1;
		}
		cwd[0] = dir[0];
		cwd[1] = dir[1];
		// Observed timing:
		// [0-562mS]
		//	(cr)
		//	Ver 03.68VDAPF On-Line:(cr)
		// [250mS]
		//	Device Detected P2(cr)
		// [16-18mS]
		//	No Upgrade(cr)
		// [1-2mS]
		//	D:\>(cr)
		// TODO: replicate real hardware timming?

		sendBack(String.format("\rVer %sF On-Line:\r", fw));
		if (dir[0] != null) {
			sendBack("Device Detected P1\r");
		}
		if (dir[1] != null) {
			sendBack("Device Detected P2\r");
		}
		sendBack("No Upgrade\r"); // TODO: probably only if USB drive inserted
		sendPrompt();
	}

	public int in(int port) {
		int off = port - basePort;
		int val = 0;
		if (off == statusPortOffset) {
			if (inBuf != null || fifi.size() > 0) {
				val |= sts_DataReady;
			}
			val |= sts_TxReady;
		} else if (off == dataPortOffset) {
			val = getBack();
		}
		return val;
	}

	public void out(int port, int val) {
		int off = port - basePort;
		if (off == dataPortOffset) {
			startLED();
			if (cmd) {
				// TODO: must be context-sensitive: number fields in IPH.
				// Must look for 0x20, then decode command, then
				// allow for number fields containing 0x0d.
				if (val == 0x0d) {
					processCmd();
				} else {
					if (cmdIdx < cmdBuf.length) {
						cmdBuf[cmdIdx++] = (byte)val;
					}
				}
			} else {
				dataBuf[dataIdx++] = (byte)val;
				if (dataIdx >= dataBuf.length) {
					finishCmd();
				}
			}
		}
	}

	private BasicFileAttributes getAttrs(String name) {
		Path path = FileSystems.getDefault().getPath(name);
		try {
			return Files.readAttributes(path, BasicFileAttributes.class);
		} catch (Exception ee) {
			return null;
		}
	}

	private long fat2unix(long fat) {
		Calendar cal = Calendar.getInstance();
		cal.setTimeZone(tz);
		int y = (int)((fat >> 25) & 0x7f) + 1980;
		int m = (int)((fat >> 21) & 0x0f) - 1;
		int d = (int)((fat >> 16) & 0x1f);
		int h = (int)((fat >> 11) & 0x1f);
		int n = (int)((fat >> 5) & 0x3f);
		int s = (int)(fat & 0x1f) * 2;
		if (m < 0 || d == 0 || m > 11 || d > 31 ||
				h > 23 || n > 59 || s > 59) {
			return -1;
		}
		cal.set(y, m, d, h, n, s);
		return cal.getTimeInMillis();
	}

	private long unix2fat(long timet) {
		Calendar cal = Calendar.getInstance();
		cal.setTimeInMillis(timet);
		cal.setTimeZone(tz);
		int y = cal.get(Calendar.YEAR) - 1980;
		int m = cal.get(Calendar.MONTH) + 1;
		int d = cal.get(Calendar.DAY_OF_MONTH);
		int h = cal.get(Calendar.HOUR_OF_DAY);
		int n = cal.get(Calendar.MINUTE);
		int s = cal.get(Calendar.SECOND) / 2;
		long fat = (((long)y & 0x7f) << 25) |
				((m & 0x0f) << 21) |
				((d & 0x1f) << 16) |
				((h & 0x1f) << 11) |
				((n & 0x3f) << 5) |
				(s & 0x1f);
		return fat;
	}

	private void fillBack() {
		if (fifi.size() > 0) {
			try {
				inBuf = fifi.take();
			} catch (Exception ee) {}
			inIdx = 0;
		} else {
			inBuf = null;
		}
	}

	private int getBack() {
		// must not sleep.
		if (inBuf == null) {
			fillBack();
			if (inBuf == null) {
				return 0;
			}
		}
		int val = inBuf[inIdx++] & 0xff;
		if (inIdx >= inBuf.length) {
			fillBack();
		}
		return val;
	}

	private void sendBack(byte[] buf) {
//System.err.format("    : \"%s\"\n", new String(buf));
		if (inBuf == null) {
			inBuf = buf;
			inIdx = 0;
			return;
		}
		fifi.add(buf);
	}

	private void sendBack(String rsp) {
		sendBack(rsp.getBytes());
	}

	private void sendError(int err) {
		if (bin) {
			sendBack(errs.get(err) + "\r");
		} else {
			sendBack(errors.get(err) + "\r");
		}
	}

	private void sendWord(long val) {
		if (hex) {
			byte[] buf = new byte[2];
			buf[0] = (byte)(val & 0xff);
			buf[1] = (byte)((val >> 8) & 0xff);
			sendBack(buf);
		} else {
			sendBack(String.format("$%02X $%02X ",
					(val & 0xff),
					((val >> 8) & 0xff)));
		}
	}

	private void sendDword(long val) {
		if (hex) {
			byte[] buf = new byte[4];
			buf[0] = (byte)(val & 0xff);
			buf[1] = (byte)((val >> 8) & 0xff);
			buf[2] = (byte)((val >> 16) & 0xff);
			buf[3] = (byte)((val >> 24) & 0xff);
			sendBack(buf);
		} else {
			sendBack(String.format("$%02X $%02X $%02X $%02X ",
					(val & 0xff),
					((val >> 8) & 0xff),
					((val >> 16) & 0xff),
					((val >> 24) & 0xff)));
		}
	}

	private void sendQword(long val) {
		if (hex) {
			byte[] buf = new byte[6];
			buf[0] = (byte)(val & 0xff);
			buf[1] = (byte)((val >> 8) & 0xff);
			buf[2] = (byte)((val >> 16) & 0xff);
			buf[3] = (byte)((val >> 24) & 0xff);
			buf[4] = (byte)((val >> 32) & 0xff);
			buf[5] = (byte)((val >> 40) & 0xff);
			sendBack(buf);
		} else {
			sendBack(String.format("$%02X $%02X $%02X $%02X $%02X $%02X",
					(val & 0xff),
					((val >> 8) & 0xff),
					((val >> 16) & 0xff),
					((val >> 24) & 0xff),
					((val >> 32) & 0xff),
					((val >> 48) & 0xff)));
		}
	}

	private int cmdFWV() {
		sendBack(String.format("\rMAIN %s\rRPRG ?.??R\r", fw));
		return 0;
	}

	// For checking names in local dirs
	private boolean ok(File ent) {
		String nm = ent.getName();
		if (nm.matches(".*[A-Z].*")) {
			return false;
		}
		return ok(nm);
	}

	private boolean ok(String nm) {
		if (nm.length() == 0) {
			return false;
		}
		int dot = nm.indexOf('.');
		if (dot == 0) {
			if (dir[idx].equals(cwd[idx])) return false;
			if (!nm.equals(".") && !nm.equals("..")) return false;
		}
		if (dot > 8) {
			return false;
		}
		if (dot < 0) {
			return (nm.length() <= 8);
		}
		if (nm.length() - dot > 4) return false;
		return true;
	}

	private long getDword() {
		// TODO: support IPH mode
		if (cmdPix >= cmdIdx) {
			return -1;
		}
		if (hex) {
			// TODO: must allow 0x0d in data!
			if (cmdIdx - cmdPix < 4) {
				return -1;
			}
			long val = ((cmdBuf[cmdPix++] & 0xff) << 24);
			val |= ((cmdBuf[cmdPix++] & 0xff) << 16);
			val |= ((cmdBuf[cmdPix++] & 0xff) << 8);
			val |= (cmdBuf[cmdPix++] & 0xff);
			return val;
		} else {
			int n = cmdPix;
			while (n < cmdIdx && cmdBuf[n] != 0x20) ++n;
			String nm = new String(cmdBuf, cmdPix, n - cmdPix)
						.replaceAll("^\\$", "0x");
			cmdPix = n;
			try {
				return Long.decode(nm);
			} catch (Exception ee) {
				return -1;
			}
		}
	}

	private File getFile() {
		if (cmdPix >= cmdIdx) {
			return null;
		}
		int n = cmdPix;
		while (n < cmdIdx && cmdBuf[n] != 0x20) ++n;
		String nm = new String(cmdBuf, cmdPix, n - cmdPix);
		cmdPix = n;
		if (!ok(nm)) {
			return null;
		}
		return new File(cwd[idx], nm.toLowerCase());
	}

	private int cmdCD() {
		if (cwd[idx] == null) {
			return ERR_ND;
		}
		if (ofile != null) {
			return ERR_FO;
		}
		++cmdPix; // skip blank
		if (cmdPix >= cmdIdx) {
			return ERR_FI;
		}
		String nm = new String(cmdBuf, cmdPix, cmdIdx - cmdPix);
		if (nm.equals("..")) {
			if (dir[idx].equals(cwd[idx]) || cwd[idx].getParentFile() == null) {
				return ERR_CF;
			}
			cwd[idx] = cwd[idx].getParentFile();
			return 0;
		}
		if (!ok(nm)) {
			return ERR_FI;
		}
		File d = new File(cwd[idx], nm.toLowerCase());
		if (!d.isDirectory()) {
			return ERR_FI;
		}
		cwd[idx] = d;
		return 0;
	}

	private int cmdDIR() {
		sendBack("\r");
		if (cwd[idx] == null) {
			return ERR_ND;
		}
		if (ofile != null) {
			return ERR_FO;
		}
		if (cmdPix < cmdIdx) {
			++cmdPix; // skip blank
			File d = getFile();
			if (d == null || !d.exists()) {
				return ERR_CF;
			}
			long l;
			try {
				l = d.length();
			} catch (Exception ee) {
				return ERR_CF;
			}
			sendBack(String.format("%s ", d.getName().toUpperCase()));
			sendDword(l);
			sendBack("\r");
			return 0;
		}
		File[] ents = cwd[idx].listFiles();
		for (File ent : ents) {
			if (!ok(ent)) {
				continue;
			}
			dir(ent);
		}
		return 0;
	}

	private void dirt(File f) {
		if (f.isDirectory()) {
			sendBack(String.format("%s DIR\r", f.getName().toUpperCase()));
		} else {
			BasicFileAttributes attrs = getAttrs(f.getAbsolutePath());
			sendBack(String.format("%s ", f.getName().toUpperCase()));
			sendDword(unix2fat(attrs.creationTime().toMillis()));
			sendWord(unix2fat(attrs.lastAccessTime().toMillis()) >> 16);
			sendDword(unix2fat(attrs.lastModifiedTime().toMillis()));
			sendBack("\r");
		}
	}

	private void dirs(File f) {
		if (f.isDirectory()) {
			sendBack(String.format("%s DIR\r", f.getName().toUpperCase()));
		} else {
			long l = 0;
			try {
				l = f.length();
			} catch (Exception ee) {}
			sendBack(String.format("%s ", f.getName().toUpperCase()));
			sendDword(l);
			sendBack("\r");
		}
	}

	private void dir(File f) {
		if (f.isDirectory()) {
			sendBack(String.format("%s DIR\r", f.getName().toUpperCase()));
		} else {
			sendBack(String.format("%s\r", f.getName().toUpperCase()));
		}
	}

	private int cmdDIRT() {
		sendBack("\r");
		if (cwd[idx] == null) {
			return ERR_ND;
		}
		if (ofile != null) {
			return ERR_FO;
		}
		++cmdPix; // skip blank
		File d = getFile();
		if (d == null) {
			return ERR_CF; // what's the right error?
		}
		if (!d.exists()) {
			return ERR_CF;
		}
		dirt(d);
		return 0;
	}

	private int cmdCLF() {
		if (cwd[idx] == null) {
			return ERR_ND;
		}
		// TODO: need to check filename?
		if (ofile != null) {
			try {
				if (ofWrite) {
					long n = ofile.getFilePointer();
					ofile.setLength(n);
				}
				ofile.close();
				// Not compelling to update access time for OPR...
				if (ofWrite && ofTime > 0) {
					// TODO: can't change create time...
					// that will always be "now".
					ofFile.setLastModified(ofTime);
				}
			} catch (Exception ee) {}
			ofile = null;
		}
		return 0;
	}

	private int cmdRD() {
		if (cwd[idx] == null) {
			return ERR_ND;
		}
		if (ofile != null) {
			return ERR_FO;
		}
		++cmdPix; // skip blank
		File d = getFile();
		if (d == null || d.isDirectory()) {
			return ERR_FI;
		}
		if (!d.exists()) {
			return ERR_CF;
		}
		try {
			RandomAccessFile f = new RandomAccessFile(d, "r");
			byte[] buf = new byte[(int)d.length()];
			f.read(buf);
			sendBack(buf);
			f.close();
		} catch (Exception ee) {
			return ERR_CF;
		}
		return 0;
	}

	private int cmdOPR() {
		if (cwd[idx] == null) {
			return ERR_ND;
		}
		if (ofile != null) {
			return ERR_FO;
		}
		++cmdPix; // skip blank
		File d = getFile();
		if (d == null || d.isDirectory()) {
			return ERR_FI;
		}
		if (!d.exists()) {
			return ERR_CF;
		}
		ofTime = -1;
		if (cmdPix < cmdIdx) { // optional timestamp
			// Not compelling to update access time for OPR...
			// plus, don't have mechanism for it.
			++cmdPix; // skip blank
			ofTime = fat2unix(getDword());
		}
		ofWrite = false;
		try {
			ofile = new RandomAccessFile(d, "r");
			ofFile = d;
		} catch (Exception ee) {
			return ERR_FI;
		}
		return 0;
	}

	private int cmdOPW() {
		if (cwd[idx] == null) {
			return ERR_ND;
		}
		if (ofile != null) {
			return ERR_FO;
		}
		++cmdPix; // skip blank
		File d = getFile();
		if (d == null || d.isDirectory()) {
			return ERR_FI;
		}
		if (d.exists() && !d.canWrite()) {
			return ERR_RO;
		}
		ofTime = -1;
		if (cmdPix < cmdIdx) { // optional timestamp
			ofTime = fat2unix(getDword());
		}
		ofWrite = true;
		try {
			ofile = new RandomAccessFile(d, "rw");
			ofFile = d;
		} catch (Exception ee) {
			return ERR_FI;
		}
		return 0;
	}

	private int cmdRDF() {
		int e = 0;
		if (cwd[idx] == null) {
			e = ERR_ND;
		} else if (ofile == null) {
			e = ERR_FI;
		}
		++cmdPix; // skip blank
		long n = getDword();
		if (n <= 0) {
			// TODO: we lose sync if we do this!
			return ERR_CF; // what's the right error?
		}
		byte[] buf = new byte[(int)n];	// zeroed?
		if (e == 0) try {
			int x = ofile.read(buf);
			if (x < n) {
				e = ERR_CF;
			}
		} catch (Exception ee) {
			e = ERR_CF; // what to do? no prescribed error code!
		}
		sendBack(buf);
		return e;
	}

	private int cmdWRF() {
		int e = 0;
		if (cwd[idx] == null) {
			e = ERR_ND;
		} else if (ofile == null || !ofWrite) {
			e = ERR_FI;
		}
		++cmdPix; // skip blank
		long n = getDword();
		if (n <= 0) {
			// TODO: we lose sync if we do this!
			return ERR_CF; // what's the right error?
		}
		dataBuf = new byte[(int)n];
		cmd = false;
		dataIdx = 0;
		cmdErr = e;
		return 0;
	}

	private int cmdSEK() {
		if (cwd[idx] == null) {
			return ERR_ND;
		} else if (ofile == null) {
			return ERR_FI;
		}
		++cmdPix; // skip blank
		long n = getDword();
		if (n < 0 || n > ofFile.length()) {
			return ERR_CF; // what's the right error?
		}
		try {
			ofile.seek(n);
		} catch (Exception ee) {
			return ERR_CF;
		}
		return 0;
	}

	private int cmdREN() {
		if (cwd[idx] == null) {
			return ERR_ND;
		} else if (ofile != null) {
			return ERR_FO;
		}
		++cmdPix; // skip blank
		File d = getFile();
		++cmdPix; // skip blank
		if (cmdPix >= cmdIdx) {
			return ERR_FI;	// what's the right error
		}
		File e = getFile();
		if (d == null || e == null) {
			return ERR_FI;
		}
		if (!d.exists() || e.exists()) {
			return ERR_CF;
		}
		try {
			boolean f = d.renameTo(e);
			if (!f) {
				return ERR_CF;
			}
		} catch (Exception ee) {
			return ERR_CF;
		}
		return 0;
	}

	private int cmdDLF() {
		if (cwd[idx] == null) {
			return ERR_ND;
		} else if (ofile != null) {
			return ERR_FO;
		}
		++cmdPix; // skip blank
		File d = getFile();
		if (d == null) {
			return ERR_FI;
		}
		if (!d.isFile()) {
			return ERR_CF;
		}
		if (!d.canWrite()) {
			return ERR_RO;
		}
		try {
			boolean e = d.delete();
			if (!e) {
				return ERR_CF;
			}
		} catch (Exception ee) {
			return ERR_CF;
		}
		return 0;
	}

	private int cmdMKD() {
		if (cwd[idx] == null) {
			return ERR_ND;
		} else if (ofile != null) {
			return ERR_FO;
		}
		++cmdPix; // skip blank
		File d = getFile();
		if (d == null) {
			return ERR_FI;
		}
		if (d.exists()) {
			return ERR_CF;
		}
		long time = -1;
		if (cmdPix < cmdIdx) { // optional timestamp
			++cmdPix; // skip blank
			time = fat2unix(getDword());
		}
		try {
			boolean e = d.mkdir();
			if (!e) {
				return ERR_CF;
			}
			if (time > 0) {
				// TODO: can't change create time...
				// that will always be "now".
				d.setLastModified(time);
			}
		} catch (Exception ee) {
			return ERR_CF;
		}
		return 0;
	}

	private int cmdDLD() {
		if (cwd[idx] == null) {
			return ERR_ND;
		} else if (ofile != null) {
			return ERR_FO;
		}
		++cmdPix; // skip blank
		File d = getFile();
		if (d == null) {
			return ERR_FI;
		}
		if (!d.isDirectory()) {
			return ERR_CF;
		}
		try {
			boolean e = d.delete();
			if (!e) {
				return ERR_NE;	// assumed
			}
		} catch (Exception ee) {
			return ERR_CF;
		}
		return 0;
	}

	// VDFC Commands:

	// Copy file A: to B:
	private int cmdCPF() {
		if (cwd[0] == null || cwd[1] == null) {
			return ERR_CF;
		}
		++cmdPix; // skip blank
		idx = 0;	// A: is source
		File s = getFile();
		if (s == null) {
			return ERR_FI;
		}
		if (!s.isFile()) {
			return ERR_CF;
		}
		File d = new File(cwd[1], s.getName());
		try {
			// TODO: also COPY_ATTRIBUTES?
			Files.copy(s.toPath(), d.toPath(),
				StandardCopyOption.REPLACE_EXISTING);
			dirs(d);
		} catch (Exception ee) {
			return ERR_CF;
		}
		return 0;
	}

	private File uniqueDir(File cwd) {
		File u;
		do {
			u = new File(cwd, String.format("vf2f%04d", uniq++));
		} while (u.exists() && uniq < 10000);
		if (u.exists()) {
			u = null;
		} else {
			try {
				if (!u.mkdir()) {
					u = null;
				}
			} catch (Exception ee) {
				u = null;
			}
		}
		return u;
	}

	// Snapshot Image file A: on B:
	private int cmdIMF() {
		if (cwd[0] == null || cwd[1] == null) {
			return ERR_CF;
		}
		++cmdPix; // skip blank
		idx = 0;	// A: is source
		File s = getFile();
		if (s == null) {
			return ERR_FI;
		}
		if (!s.isFile()) {
			return ERR_CF;
		}
		File d = uniqueDir(cwd[1]);
		if (d == null) {
			// TODO: what's the right handling?
			return ERR_CF;
		}
		sendBack(String.format("%s\r", d.getName().toUpperCase()));
		d = new File(d, s.getName());
		try {
			// TODO: also COPY_ATTRIBUTES?
			Files.copy(s.toPath(), d.toPath(),
				StandardCopyOption.REPLACE_EXISTING);
			dirs(d);
		} catch (Exception ee) {
			return ERR_CF;
		}
		return 0;
	}

	private class CopyFiles extends SimpleFileVisitor<Path> {
		File cd;
		public CopyFiles(File start) {
			cd = start;
		}
		@Override
		public FileVisitResult visitFile(Path file,
				BasicFileAttributes attr) {
			File d = new File(cd, file.toFile().getName().toUpperCase());
			try {
				Files.copy(file, d.toPath(),
					StandardCopyOption.REPLACE_EXISTING);
			} catch (Exception ee) {
				return FileVisitResult.TERMINATE;
			}
			dirs(d);
			return FileVisitResult.CONTINUE;
		}
		@Override
		public FileVisitResult preVisitDirectory(Path file,
				BasicFileAttributes attr) {
			if (cwd[0].equals(file.toFile())) {
				return FileVisitResult.CONTINUE;
			}
			cd = new File(cd, file.toFile().getName());
			try {
				if (!cd.mkdir()) {
					return FileVisitResult.TERMINATE;
				}
			} catch (Exception ee) {
				return FileVisitResult.TERMINATE;
			}
			dir(file.toFile());
			return FileVisitResult.CONTINUE;
		}
		@Override
		public FileVisitResult postVisitDirectory(Path file,
				IOException ee) {
			if (cwd[0].equals(file.toFile())) {
				return FileVisitResult.CONTINUE;
			}
			cd = cd.getParentFile();
			sendBack(".. DIR\r");
			// TODO: can it ever be null?
			return FileVisitResult.CONTINUE;
		}
	}

	// Copy subtree A:(CWD) to B:(CWD) (recursive)
	private int cmdCPS() {
		if (cwd[0] == null || cwd[1] == null) {
			return ERR_CF;
		}
		idx = 0;
		try {
			Files.walkFileTree(cwd[0].toPath(), new CopyFiles(cwd[1]));
		} catch (Exception ee) {
			return ERR_CF;
		}
		return 0;
	}

	// Snapshot Image A:(CWD) on B:(CWD) (recursive)
	private int cmdIMS() {
		if (cwd[0] == null || cwd[1] == null) {
			return ERR_CF;
		}
		File start = uniqueDir(cwd[1]);
		if (start == null) {
			// TODO: what's the right handling?
			return ERR_CF;
		}
		sendBack(String.format("%s\r", start.getName().toUpperCase()));
		try {
			Files.walkFileTree(cwd[0].toPath(), new CopyFiles(start));
		} catch (Exception ee) {
			return ERR_CF;
		}
		idx = 0;
		return 0;
	}

//private static void copyFileUsingJava7Files(File source, File dest) throws IOException {
//    Files.copy(source.toPath(), dest.toPath());
//}

	private void processCmd() {
		int c, n;
//System.err.format("VDIP1: \"%s\" (\"%s\"/%d)\n", new String(cmdBuf, 0, cmdIdx),
//				(inBuf != null ? new String(inBuf) : ""),
//				(inBuf != null ? inIdx : -1));
//System.err.format("VDIP1:");
//for (n = 0; n < cmdIdx; ++n) {
//System.err.format(" %02x", cmdBuf[n] & 0xff);
//}
//System.err.format("\n");
		cmdPix = 0;
		if (cmdIdx == 0) {
			c = 0;
		} else if (bin) {
			c = cmdBuf[cmdPix] & 0xff; // could be 0x20 (' ')
			++cmdPix;
			if (c == 0x41 || c == 0x42) { // kludge...
				c = (c << 8) | (cmdBuf[cmdPix] & 0xff);
				++cmdPix;
			}
		} else {
			n = 0;
			while (n < cmdIdx && cmdBuf[n] != 0x20) ++n;
			String cm;
			if (n == 1) {
				// assume E/e is the only one char cmd,
				// preserve case.
				cm = new String(cmdBuf, 0, n);
			} else {
				cm = new String(cmdBuf, 0, n).toUpperCase();
			}
			Integer cc = cmds.get(cm);
			if (cc == null) {
				sendError(ERR_BC);
				cmdIdx = 0;
				return;
			}
			c = cc;
			cmdPix = n;
		}
		int err = ERR_BC;
		switch(c) {
		case 0: // empty command
			err = (dir == null ? ERR_ND : 0);
			break;
		case CMD_SBD:
			err = 0; // passively accept command
			break;
		case CMD_FWV:
			err = cmdFWV();
			break;
		case CMD_FS:
			sendDword(0xffffffffL);
			sendBack("\r");
			err = 0;
			break;
		case CMD_FSE:
			sendQword(64L*1024*1024*1024);
			sendBack("\r");
			err = 0;
			break;
		case CMD_DIR:
			err = cmdDIR();
			break;
		case CMD_DIRT:
			err = cmdDIRT();
			break;
		case CMD_CD:
			err = cmdCD();
			break;
		case CMD_SCS:
		case CMD_ECS:
			err = 0;
			bin = (c == CMD_SCS);
			break;
		case CMD_IPA:
		case CMD_IPH:
			err = 0;
			hex = (c == CMD_IPH);
			break;
		case CMD_E:
		case CMD_e:
			err = 0;
			sendBack(String.format("%c\r", c));
			cmdIdx = 0;
			return;
		case CMD_CLF:
			err = cmdCLF();
			break;
		case CMD_RD:
			err = cmdRD();
			break;
		case CMD_OPR:
			err = cmdOPR();
			break;
		case CMD_OPW:
			err = cmdOPW();
			break;
		case CMD_RDF:
			err = cmdRDF();
			break;
		case CMD_WRF:
			err = cmdWRF();
			if (err == 0) {
				return;
			}
			break;
		case CMD_SEK:
			err = cmdSEK();
			break;
		case CMD_DLF:
			err = cmdDLF();
			break;
		case CMD_MKD:
			err = cmdMKD();
			break;
		case CMD_DLD:
			err = cmdDLD();
			break;
		case CMD_REN:
			err = cmdREN();
			break;
		// VDFC commands:
		case CMD_A:
		case CMD_B:
			if (!vdfc) break;	// ERR_BC
			n = (c >> 8) - 'A';
			if (dir[n] == null) {
				err = ERR_CF;
			} else {
				err = 0;
				idx = n;
			}
			break;
		case CMD_CPF:
			if (!vdfc) break;	// ERR_BC
			err = cmdCPF();
			break;
		case CMD_IMF:
			if (!vdfc) break;	// ERR_BC
			err = cmdIMF();
			break;
		case CMD_CPS:
			if (!vdfc) break;	// ERR_BC
			err = cmdCPS();
			break;
		case CMD_IMS:
			if (!vdfc) break;	// ERR_BC
			err = cmdIMS();
			break;
		default:
			break;
		}
		if (err > 0) {
			sendError(err);
		} else {
			sendPrompt();
		}
		cmdIdx = 0;
	}

	private void finishCmd() {
		// Only WRF?
		if (cmdErr == 0) try {
			ofile.write(dataBuf);
		} catch (Exception ee) {
			cmdErr = ERR_CF;
		}
		cmd = true;
		dataBuf = null;	// free up buffer
		if (cmdErr > 0) {
			sendError(cmdErr);
		} else {
			sendPrompt();
		}
		cmdIdx = 0;
	}

	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == timer) {
			timer.removeActionListener(this);
			if (led != null) {
				led.set(false);
			}
			ledState = false;
			return;
		}
		if (!(e.getSource() instanceof JMenuItem)) {
			return;
		}
		if (e.getActionCommand().equals("menu")) {
			//mnt = (JMenuItem)e.getSource();
			return;
		}
		//System.err.format("VDIP1 Menu\n");
		if (vdfc) {
			JMenuItem mi = (JMenuItem)e.getSource();
			int k = mi.getMnemonic();
			if (k == KeyEvent.VK_P) {
				pu.show(led, 0, 0);
			} else if (k == KeyEvent.VK_1) {
				chooseDir(0);
			} else if (k == KeyEvent.VK_2) {
				chooseDir(1);
			}
		} else {
			chooseDir(idx);
		}
	}

	private void chooseDir(int x) {
		if (dir[x] != null) {
			sendBack(String.format("Device Removed P%d\r", x + 1));
		}
		SuffFileChooser sfc = new SuffFileChooser("USBDrive Dir",
			null, null, null, null, dir[x], 0);
		sfc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
		int rv = sfc.showDialog(null);
		if (rv == JFileChooser.APPROVE_OPTION) {
			setDir(x, sfc.getSelectedFile());
			sendBack(String.format("Device Detected P%d\r", x + 1));
		} else {
			setDir(x, null);
		}
	}

	public String dumpDebug() {
		String ret = new String();
		ret += String.format("port %02x", basePort);
		ret += String.format("\ninBuf: %d, %d \"%s\" (#fifi = %d)",
				(inBuf != null ? inBuf.length : 0), inIdx,
				(inBuf != null ? new String(inBuf) : ""),
				fifi.size());
		if (dir[0] != null) {
			ret += String.format("\ndir1 = %s", dir[0].getAbsolutePath());
			ret += String.format("\ncwd1 = %s", cwd[0].getAbsolutePath());
		}
		if (dir[1] != null) {
			ret += String.format("\ndir2 = %s", dir[1].getAbsolutePath());
			ret += String.format("\ncwd2 = %s", cwd[1].getAbsolutePath());
		}
		if (ofile != null) {
			long fp = 0;
			ret += String.format("\nop%s: %s", ofWrite ? "w" : "r",
					ofFile.getAbsolutePath());
			try {
				fp = ofile.getFilePointer();
			} catch (Exception ee) {}
			ret += String.format("\n     ptr = %d", fp);
		}
		return ret + '\n';
	}

}
