// Copyright (c) 2016 Douglas Miller <durgadas311@gmail.com>

import java.util.Properties;
import java.io.*;

public class X2H8Memory extends H89Roms implements IODevice, Memory, GppListener {
	int base;
	boolean rom;
	GeneralPurposePort gpp;
	int gpioBit;
	int gpioORG0;
	boolean fault;
	// Organized as [16][16384]...
	private byte[] mem;
	private int curBank[];
	private byte[] ff;

	public X2H8Memory(Properties props, int port,
			GeneralPurposePort gpio, Interruptor intr) {
		super(props, gpio, intr);
		gpp = gpio;
		base = port;
		String s = props.getProperty("x2h8_port");
		if (s != null) {
			base = Integer.decode(s) & 0xff;
		}
		s = props.getProperty("x2h8_fault");
		fault = (s != null);
		mem = new byte[262144]; // 16x16K = 256K
		ff = new byte[4];
		curBank = new int[4]; // 4 - 16K pages in 64K addr space
		reset();
		gpioORG0 = gpp.getOrg0Bit();
		gpioBit = romBit | gpioORG0;
		gpp.addGppListener(this);
		System.err.format("X/2-H8 Memory Controller at port %02x\n", base);
	}

	public int getBaseAddress() { return base; }
	public int getNumPorts() { return 1; }
	public int in(int port) {
		return 0xff;
	}
	// TODO: replicate pathological behavior when multiple banks selected
	public void out(int port, int value) {
		int b = (value >> 4) & 0x03;
		value &= 0x0f;
		int n = b << 16; // base of bank
		int d = ff[b] ^ value;
		ff[b] = (byte)value;
		if ((value & 0x01) == 0) curBank[0] = n;
		n += 16384;
		if ((value & 0x02) == 0) curBank[1] = n;
		n += 16384;
		if ((value & 0x04) == 0) curBank[2] = n;
		n += 16384;
		if ((value & 0x08) == 0) curBank[3] = n;
		if (!fault) {
			return;
		}
		// detect segments not selected at all
		d = ff[0] & ff[1] & ff[2] & ff[3];
		if ((d & 0x01) != 0) curBank[0] = -1;
		if ((d & 0x02) != 0) curBank[1] = -1;
		if ((d & 0x04) != 0) curBank[2] = -1;
		if ((d & 0x08) != 0) curBank[3] = -1;
		// detect segments with multiple selections
		d = (~ff[0] & 1) + (~ff[1] & 1) + (~ff[2] & 1) + (~ff[3] & 1);
		if (d != 1) curBank[0] = -1;
		d = (~ff[0] & 2) + (~ff[1] & 2) + (~ff[2] & 2) + (~ff[3] & 2);
		if (d != 2) curBank[1] = -1;
		d = (~ff[0] & 4) + (~ff[1] & 4) + (~ff[2] & 4) + (~ff[3] & 4);
		if (d != 4) curBank[2] = -1;
		d = (~ff[0] & 8) + (~ff[1] & 8) + (~ff[2] & 8) + (~ff[3] & 8);
		if (d != 8) curBank[3] = -1;
	}
	public String getDeviceName() { return "X2-H8"; }
	public String dumpDebug() {
		String ret = String.format("ROM: %s\n", rom);
		ret += String.format("PAGE 0000 = %05X\n", curBank[0]);
		ret += String.format("PAGE 4000 = %05X\n", curBank[1]);
		ret += String.format("PAGE 8000 = %05X\n", curBank[2]);
		ret += String.format("PAGE C000 = %05X\n", curBank[3]);
		ret += String.format("FF: %x %x %x %x\n", ff[0], ff[1], ff[2], ff[3]);
		ret += String.format("Fault detect: %s\n", fault);
		ret += '\n';
		ret += mon.dumpDebug();
		if (h17 != null) ret += h17.dumpDebug();
		return ret;
	}

	private int read(boolean rom, int address) {
		if (rom && address < romMask) {
			if (address < monMask || (h17 == null && address >= 0x1800)) {
				return mon.read(address);
			}
			if (address >= 0x1800 && h17 != null) {
				return h17.read(address - 0x1800);
			}
		}
		return mem[address] & 0xff;
	}
	public int read(boolean rom, int bank, int address) {
		return read(rom, (bank << 16) | address);
	}
	public int read(int address) {
		int pg = curBank[(address >> 14) & 0x0003];
		if (pg < 0) {
			return 0;
		}
		address &= 0x3fff;
		return read(rom, pg + address);
	}
	public void write(int address, int value) {
		if (rom && address < monMask) mon.write(address, value);
		int pg = curBank[(address >> 14) & 0x0003];
		if (pg < 0) {
			return;
		}
		address &= 0x3fff;
		mem[pg + address] = (byte)value;
	}

	public void reset() {
		rom = true;
		ff[0] = (byte)0;
		ff[1] = ff[2] = ff[3] = (byte)0x0f;
		out(base, 0x00);
	}

	public void dumpCore(String file) {
		try {
			OutputStream core = new FileOutputStream(file);
			core.write(mem);
			core.close();
		} catch (Exception ee) {
			H89Operator.error(null, "Core Dump", ee.getMessage());
		}
	}

	public int interestedBits() { return gpioBit; }
	public void gppNewValue(int gpio) {
		rom = ((gpio & gpioORG0) == 0);
		if (romBit != 0) romGpp(gpio);
	}
}
