// Copyright (c) 2016 Douglas Miller <durgadas311@gmail.com>

import java.util.Arrays;
import java.util.Vector;
import java.util.Properties;

//
// State Flow Diagram:
//
//                                         IDLE
//                           (sync)         |                       (write-gate)
//                             +------------+-----------------------------+
//     (prev FORMAT)           |                                          |
//             +---------------+                                          |
//             |               |                                          |
//    +-------------------> HDR_SYNC                                   FORMAT <---+
//    |        |               |                                          |       |
//    |     FMT_SYNC         HEADER                                       |       |
//    |        |        (hole) |        (sync)        (write-gate)        |       |
//    |     FORMAT        +----+-----------+--------------+               |       |
//    |        |          |                |              |               |       |
//    |        |          |            DATA_SYNC         DATA      [write format] |
//    |        |         IDLE              |              |               |       |
//    |  [read format]                   DATA        [write sector]       |       |
//    |        |                           |              |               |**     |
//    |        |                     [read sector]        |*              |       |
//    | (sync) |  (write-gate)  (T)        |              |               |       |
//    |   +----+-----+-----------+     (end data)      (wg end)        (wg end)   |
//    |   |          |           +---------++-------------+---------------+       |
//    +---+          |                      |                                     |
//                   +--------------------- | ------------------------------------+
//                                         IDLE
// *  Discard past end-of-data
// ** Wrap around track buffer and end
// (T) = Terminal event: WRITE, STEP, DESELECT, MOTOR.
// SYNC states consist of zero-or-more sync-00 followed by a single sync-FD.
//
// Sector holes are 20mS apart, about 320 bytes. Index hole is 10mS (160 bytes)
// from adjacent sector holes. The index hole is in the middle of sector data,
// so if in any of HDR_SYNC, HEADER, or DATA_SYNC when a hole passes by,
// we are in an error case and need to abort.

public class Z17 implements DiskController, ClockListener, GppListener {
	static final int BasePort_c = 0x7c;
	static final int Z17_NumPorts_c = 4;
	static final int ControlPort_Offset_c = 3; // and Disk Status
	static final int RcvCtrlPort_Offset_c = 2; // out char, input if "found"...
	static final int StatusPort_Offset_c = 1; // USART status
	static final int DataPort_Offset_c = 0;
	static final int numDisks_c = 3;
	static final int cyclesPerByte_c = 128;
	static final boolean DD = false; // never supports doouble density
	static final int syncValue_c = 0xfd;

	static final int ctrl_WriteGate_c = 0x01;
	static final int ctrl_DS0_c = 0x02;
	static final int ctrl_DS1_c = 0x04;
	static final int ctrl_DS2_c = 0x08;
	static final int ctrl_DriveSel_c = 0x0e;
	static final int ctrl_DriveSelShift_c = 1;
	static final int ctrl_MotorOn_c = 0x10;
	static final int ctrl_Dir_c = 0x20;
	static final int ctrl_Step_c = 0x40;
	static final int ctrl_FMWEN_c = 0x80;

	static final int csts_IndexPulse_c = 0x01;
	static final int csts_TrackZero_c = 0x02;
	static final int csts_WriteProt_c = 0x04;
	static final int csts_Sync_c = 0x08;

	static final int sts_TxEmpty_c = 0x80;
	static final int sts_FillCharTx_c = 0x40; // required by CONFIGUR.COM
	static final int sts_Sync_c = 0x08;
	static final int sts_ParityErr_c = 0x04;
	static final int sts_Overrun_c = 0x02;
	static final int sts_RxReady_c = 0x01;
	static final int RR_reset = (sts_RxReady_c | sts_FillCharTx_c | sts_Sync_c |
			sts_ParityErr_c | sts_Overrun_c);

	private String Z17_Name = "Z17";
	private int h89_gppSideSelect;

	private GenericFloppyDrive[] drives_m;
	private LED[] leds_m;

	private int controlReg_m;
	private int fillReg_m;
	private int syncReg_m;
	private int rcvrReg_m;
	private int dataReg_m;
	private boolean rcvrSearch;
	private boolean SCRLatch;
	private boolean WG;

	private int curDriveIndex; // convenience decode of bitmap
	private GenericFloppyDrive curDrive = null;
	private int crc = 0;
	private int side = 0;
	private enum States { IDLE, HEADER, PRE_DATA, DATA, FORMAT };
	private States state = States.IDLE;
	private States prevState = States.IDLE;
	private long lastPos;
	private int sectorPos_m;
	private int writePos_m;
	private int readPos_m;
	private int indexCount;
	private int tdataCount;
	private int fillCount;
	private int[] adr;
	private byte[] trk = null;

	public Z17(Properties props, String slot, LEDHandler lh,
			Interruptor intr, GeneralPurposePort gpio) {
		h89_gppSideSelect = gpio.getSideSelBit();
		intr.addClockListener(this);
		gpio.addGppListener(this);
		drives_m = new GenericFloppyDrive[numDisks_c];
		leds_m = new LED[numDisks_c];
		String s;
		Arrays.fill(drives_m, null);
		controlReg_m = 0;

		s = props.getProperty(slot);
		String[] ss = s.split("\\s");
		Z17_Name = ss[0];
		String pfx = Z17_Name.toLowerCase();

		// First identify what drives are installed.
		GenericFloppyDrive drv;
		for (int x = 0; x < numDisks_c; ++x) {
			String prop = String.format(pfx + "_drive%d", x + 1);
			s = props.getProperty(prop);
			if (s != null) {
				drv = installDrive(x, s, lh);
			}
		}

		// Next identify what diskettes are pre-inserted.
		for (int x = 0; x < numDisks_c; ++x) {
			String prop = String.format(pfx + "_disk%d", x + 1);
			s = props.getProperty(prop);
			if (s != null) {
				drv = getDrive(x);
				if (drv != null) {
					drv.insertMedia(s.split("\\s"));
				}
			}
		}
		System.err.format("%s at port %02x\n", Z17_Name, BasePort_c);
		reset();
	}

	private GenericFloppyDrive installDrive(int x, String s, LEDHandler lh) {
		String[] ss = s.split("\\s", 2);
		String n = getDriveName(x);
		if (ss.length >= 2) {
			n = ss[1];
		}
		GenericFloppyDrive drv = GenericFloppyDrive.getInstance(ss[0], n, lh);
		if (!connectDrive(x, drv)) {
			return null;
		}
		leds_m[x] = drv.getLED();
		return drv;
	}

	public int getBaseAddress() { return BasePort_c; }
	public int getNumPorts() { return Z17_NumPorts_c; }

	private void setState(States newst) {
		if (newst == States.FORMAT && trk == null) {
			trk = new byte[3200];
		}
		if (state == States.FORMAT) {
			// debuging
		}
		if (state != States.IDLE) {
			prevState = state;
		}
		state = newst;
	}

	public void setCurLed(boolean on) {
		// This could be null
		if (curDrive == null) {
			return;
		}
		if (leds_m[curDriveIndex] != null) {
			leds_m[curDriveIndex].set(on);
		}
	}

	public GenericFloppyDrive getCurDrive() {
		return curDrive;
	}

	public Vector<GenericRemovableDrive> getDiskDrives() {
		Vector<GenericRemovableDrive> drives = new Vector<GenericRemovableDrive>();
		for (int x = 0; x < numDisks_c; ++x) {
			drives.add(drives_m[x]);	// might be null, but must preserve number.
		}
		return drives;
	}

	public String getDriveName(int index) {
		if (index < 0 || index >= numDisks_c) {
			return null;
		}
		String str = String.format("%s-%d", Z17_Name, index + 1);
		return str;
	}

	public String getDeviceName() { return Z17_Name; }

	public GenericRemovableDrive findDrive(String ident) {
		for (int x = 0; x < numDisks_c; ++x) {
			if (drives_m[x] == null) {
				continue;
			}
			if (ident.equals(drives_m[x].getDriveName())) {
				return drives_m[x];
			}
		}
		return null;
	}

	public void destroy() {
		for (int x = 0; x < numDisks_c; ++x) {
			if (drives_m[x] != null) {
				drives_m[x].insertMedia(null);
			}
		}
	}

	// NOTE: this routine is called by the same thread that calls in()/out(),
	// so there are no threading issues. The next in()/out() should see the
	// new state.
	public void addTicks(int cycleCount, long clk) {
		if (curDrive != null) {
			boolean ix = curDrive.getIndexPulse();
			curDrive.addTicks(cycleCount, clk);
			if (!ix && curDrive.getIndexPulse()) {
				adr = curDrive.readAddress();
				// NOTE: it is valid to start a HDR_SYNC before the index pulse...
				if (state == States.HEADER) {
					// if we're not in sector data by now, reset...
					// we pass through here while searching for a sector
					setState(States.IDLE);
				}
			}
			long pos = curDrive.getSectorPos(DD);
			if (!WG && pos != lastPos) {
				// TODO: check for skip?
				int db = getDataReg(pos);
				if (db == syncReg_m) {
					rcvrReg_m |= sts_Sync_c;
					SCRLatch = true;
				} else {
					rcvrReg_m &= ~sts_Sync_c;
				}
				if (rcvrSearch && SCRLatch) {
					rcvrSearch = false;
				}
				if (!rcvrSearch) {
					dataReg_m = db;
					if ((rcvrReg_m & sts_RxReady_c) != 0) {
						rcvrReg_m |= sts_Overrun_c;
					} else {
						// sts_RxReady_c going high now...
						// According to datasheet, this resets ROR
						rcvrReg_m &= ~sts_Overrun_c;
					}
					rcvrReg_m |= sts_RxReady_c;
				}
			}
			lastPos = pos;
		}
		if (fillCount > 0) {
			fillCount -= cycleCount;
			if (fillCount <= 0) {
				rcvrReg_m |= sts_FillCharTx_c;
			}
		}
		if (tdataCount > 0) {
			tdataCount -= cycleCount;
			if (tdataCount <= 0) {
				rcvrReg_m |= sts_TxEmpty_c;
				fillCount = cyclesPerByte_c;
			}
		}
	}

	private void setControlReg(int val) {
		int diff = controlReg_m ^ val;
		controlReg_m = val;
		if ((diff & ctrl_WriteGate_c) != 0) {
			WG = ((controlReg_m & ctrl_WriteGate_c) != 0);
			if (!WG) {
				// WRITE_GATE off, finish up...
				if (state == States.FORMAT) {
					Arrays.fill(trk, writePos_m, trk.length, (byte)fillReg_m);
				} else if (state == States.DATA) {
					// TODO: fill end of sector?
				}
				setState(States.IDLE);
			} else if (state == States.IDLE || state == States.FORMAT) {
				// WRITE_GATE on, FORMAT setup...
				setState(States.FORMAT);
				writePos_m = 0;
				indexCount = 0;
			} else if (state == States.HEADER) {
				// WRITE_GATE on, write sector setup...
				setState(States.PRE_DATA);
			}
		}
		if ((diff & ctrl_DriveSel_c) != 0) {
			if (curDrive != null) {
				curDrive.motor(false);
			}
			setCurLed(false);
			int dsx = (controlReg_m & ctrl_DriveSel_c) >> ctrl_DriveSelShift_c;
			curDriveIndex = Integer.numberOfTrailingZeros(dsx);
			if (curDriveIndex < drives_m.length) {
				curDrive = drives_m[curDriveIndex];
				if (curDrive != null) {
					curDrive.selectSide(side);
					curDrive.motor((controlReg_m & ctrl_MotorOn_c) != 0);
					// TODO: anything for write gate?
				}
			} else {
				curDrive = null;
			}
			setCurLed(true);
			setState(States.IDLE);
			prevState = States.IDLE;
		} else if (curDrive != null) {
			if ((diff & ctrl_MotorOn_c) != 0) {
				curDrive.motor((controlReg_m & ctrl_MotorOn_c) != 0);
				setState(States.IDLE);
				prevState = States.IDLE;
			}
			if ((diff & ctrl_Step_c) != 0 &&
					(controlReg_m & ctrl_Step_c) != 0) {
				curDrive.step((controlReg_m & ctrl_Dir_c) != 0);
				setState(States.IDLE);
				prevState = States.IDLE;
			}
		}
	}

	// Must have no side-affects.
	private int getStatusReg() {
		int val = 0;
		if (curDrive == null) {
			return val;
		}
		if (curDrive.getIndexPulse()) {
			val |= csts_IndexPulse_c;
		}
		if (curDrive.getTrackZero()) {
			val |= csts_TrackZero_c;
		}
		if (curDrive.isWriteProtect()) {
			val |= csts_WriteProt_c;
		}
		if (SCRLatch) {
			val |= csts_Sync_c;
		}
		return val;
	}

	private int getFormatByte() {
		int val = 0;
		if (trk != null) {
			val = trk[readPos_m++] & 0xff;
			if (readPos_m >= trk.length) {
				readPos_m = 0;
				++indexCount;
			}
		}
		return val;
	}

	private int getFormatSync() {
		int val = 0;
		if (trk != null) {
			int x;
			for (x = 0; x < trk.length &&
				(trk[x] & 0xff) != syncReg_m; ++x);
			if (x < trk.length) {
				readPos_m = x;
			}
			val = getFormatByte();
		}
		return val;
	}

	// This is called asynchronously when the receiver shift register fills...
	// never called if curDrive == null.
	private int getDataReg(long pos) {
		int val = 0;
		boolean doCrc = false;
		switch(state) {
		case IDLE:
			if (prevState == States.FORMAT && !SCRLatch) {
				// If after a test format, might sync from anywhere.
				readPos_m = 0;
				setState(States.FORMAT);
				val = getFormatSync();
				indexCount = 0;
			} else if (pos == 11) {
				// auto transition to HEADER...
				readPos_m = 0;
				setState(States.HEADER);
				val = syncValue_c;
				crc = 0;
			}
			break;
		case FORMAT: // This means read format
			// already did SYNC... ?
			val = getFormatByte();
			break;
		case HEADER:
			// TODO: reset CRC before this...
			switch(readPos_m) {
			case 0:
				doCrc = true;
				val = adr[2]; // side
				break;
			case 1:
				doCrc = true;
				val = adr[0]; // track
				break;
			case 2:
				doCrc = true;
				val = adr[1]; // sector
				break;
			case 3:
				// might switch to WRITE at this point...
				// stay in HEADER state, keep returning 00...
				val = crc;
				break;
			default:
				if (pos == 31) {
					// auto transition to DATA...
					readPos_m = -1; // for readData(), and ++
					int data = curDrive.readData(false, 0, 0, 0xf0, readPos_m);
					if (data != GenericFloppyFormat.DATA_AM) {
						setState(States.IDLE);
						// TODO: how to indicate errors?
					} else {
						setState(States.DATA);
						val = syncValue_c;
					}
					crc = 0;
				}
				// TODO: try to locate current sector?
				break;
			}
			++readPos_m;
			break;
		case DATA:
			int data = curDrive.readData(false, 0, 0, 0xf0, readPos_m);
			if (data >= 0) {
				doCrc = true;
				val = data;
			} else if (data == GenericFloppyFormat.CRC) {
				// Normal end of sector data...
				setState(States.IDLE);
				val = crc;
			} else {
				// GenericFloppyFormat.NO_DATA
				// GenericFloppyFormat.ID_AM
				// data < 0
				setState(States.IDLE);
				break;
			}
			++readPos_m;
			break;
		default:
			// Could use state and readPos_m to return gaps/sync.
			break;
		}
		if (doCrc) {
			crc ^= val;
			crc <<= 1;
			crc |= (crc >> 8);
			crc &= 0xff;
		}
		return val;
	}

	// WRITE-GATE logic should have prepared the state...
	private void setDataReg(int val) {
		int result;
		dataReg_m = val;
		// TODO: set rcvr status?
		if (curDrive == null) {
			return;
		}
		switch(state) {
		case IDLE:
			// passively ignore
			break;
		case FORMAT: // This means formatting disk
			trk[writePos_m++] = (byte)val;
			if (writePos_m >= trk.length) {
				writePos_m = 0;
				++indexCount;
			}
			break;
		case PRE_DATA:
			// Can we guarantee that syncReg_m has the correct value?
			if (val == syncValue_c) {
				writePos_m = 0;
				result = curDrive.writeData(false, 0, 0, 0xf0, -1, 0, true);
				if (result == GenericFloppyFormat.DATA_AM) {
					setState(States.DATA);
				} else {
					setState(States.IDLE);
					// TODO: how to indicate errors?
				}
			}
			break;
		case DATA:
			// must discard crc byte... should just return CRC (again)
			result = curDrive.writeData(false, 0, 0, 0xf0, writePos_m, val, true);
			if (result >= 0) {
				++writePos_m;
			} else if (result == GenericFloppyFormat.CRC) {
				// normal end of sector write...
				// TODO: get CRC byte and save it...
				setState(States.IDLE);
				// nothing to do, no CRC needs to be written.
			} else {
				// GenericFloppyFormat.NO_DATA
				// GenericFloppyFormat.INDEX_AM
				// GenericFloppyFormat.DATA_AM
				// result < 0
				setState(States.IDLE);
			}
			break;
		default:
			// We hit here when padding sector with NULs...
			// It is OK to just discard those bytes.
			break;
		}
	}

	public void reset() {
		fillReg_m = 0;
		setControlReg(0);
		WG = false;
		SCRLatch = false;
		rcvrSearch = true;
		setState(States.IDLE);
		tdataCount = 0;
		fillCount = cyclesPerByte_c;
		rcvrReg_m = sts_TxEmpty_c;
		fillReg_m = 0;
		syncReg_m = 0;
		dataReg_m = 0;
		lastPos = -1;
	}

	public int in(int addr) {
		int offset = addr - BasePort_c;
		int val = 0;
		if (offset == ControlPort_Offset_c) {
			val = getStatusReg();
		} else if (offset == RcvCtrlPort_Offset_c) {
			// RR
			rcvrReg_m &= ~RR_reset;
			rcvrSearch = true;
			if (tdataCount <= 0) {
				fillCount = 1;
			}
			SCRLatch = false;
		} else if (offset == StatusPort_Offset_c) {
			// SWE
			val = rcvrReg_m;
			rcvrReg_m &= ~RR_reset;
			if (tdataCount <= 0) {
				fillCount = 1;
			}
		} else if (offset == DataPort_Offset_c) {
			// RDE
			val = dataReg_m;
			rcvrReg_m &= ~sts_RxReady_c;
		} else {
			System.err.format("Z17::in(Unknown - 0x%02x)\n", addr);
		}
		return val;
	}

	public void out(int addr, int val) {
		int offset = addr - BasePort_c;
		if (offset == ControlPort_Offset_c) {
			setControlReg(val);
		} else if (offset == RcvCtrlPort_Offset_c) {
			// RSS
			syncReg_m = val; // might need to echo this back
		} else if (offset == StatusPort_Offset_c) {
			// TFS
			// Set fill character...
			fillReg_m = val; // does this ever need to be written?
		} else if (offset == DataPort_Offset_c) {
			// TDS
			setDataReg(val);
			rcvrReg_m &= ~sts_TxEmpty_c;
			tdataCount = cyclesPerByte_c;
			fillCount = 0;
		} else {
			System.err.format("Z17::out(Unknown - 0x%02x): %d\n", addr, val);
		}
	}

	public GenericFloppyDrive getDrive(int unitNum) {
		if (unitNum < numDisks_c) {
			return drives_m[unitNum];
		}
		return null;
	}

	public boolean connectDrive(int unitNum, GenericFloppyDrive drive) {
		if (unitNum >= numDisks_c) {
			System.err.format("Z17 Invalid unit number (%d)\n", unitNum);
			return false;
		}
		if (drives_m[unitNum] != null) {
			System.err.format("Z17 %d: drive already connect\n", unitNum);
			return false;
		}
		if (drive.getMediaSize() != 5) {
			System.err.format("Z17 %d: drive incompatible\n", unitNum);
			return false;
		}
		drives_m[unitNum] = drive;
		return true;
	}

	public boolean removeDrive(int unitNum) {
		return false;
	}

	public int interestedBits() {
		return h89_gppSideSelect;
	}
	public void gppNewValue(int gpio) {
		side = (gpio & h89_gppSideSelect) == 0 ? 0 : 1;
		if (curDrive != null) {
			curDrive.selectSide(side);
		}
	}

	public String dumpDebug() {
		String ret = String.format("CTRL=%02x state %s pos=%d\n" +
			"Rcvr STS %02x",
			controlReg_m, state.name(), sectorPos_m,
			rcvrReg_m);
		return ret;
	}
}
